"""
Required Notice: Copyright (C) Zoomer Analytics GmbH.

xlwings PRO is dual-licensed under one of the following licenses:

* PolyForm Noncommercial License 1.0.0 (for noncommercial use):
  https://polyformproject.org/licenses/noncommercial/1.0.0
* xlwings PRO License (for commercial use):
  https://github.com/xlwings/xlwings/blob/main/LICENSE_PRO.txt

Commercial licenses can be purchased at https://www.xlwings.org
"""

import hashlib
import re
import socket
import sys
from functools import lru_cache
from pathlib import Path

try:
    import requests
except ImportError:
    requests = None

from .. import XlwingsError
from ..utils import read_user_config
from .utils import LicenseHandler


@lru_cache(None)
def verify_execute_permission(command=None, module_names=None):
    LicenseHandler.validate_license("permissioning")
    if command:
        assert not module_names
        if re.compile(r"from .* import .*").search(command):
            raise XlwingsError("Can't verify 'from x import y' imports.")
        module_names = re.findall(r"import ([^;]*)", command)
    elif module_names:
        assert not command
    else:
        raise ValueError("You must either provide command or module_names!")
    file_names = [module + ".py" for module in module_names]
    file_hashes = {}
    for fn in file_names:
        for path in sys.path:
            # Can't use pkgutil or importlib as they may import, i.e. run the module
            if (Path(path) / fn).is_file():
                with open(Path(path) / fn, "rb") as f:
                    content = f.read()
                file_hashes[fn] = hashlib.sha256(content).hexdigest()
                break
        if fn not in file_hashes:
            raise FileNotFoundError(f"Couldn't find {fn}")

    config = read_user_config()
    method = config.get("permission_check_method", "GET").upper()

    if method == "GET":
        response = requests.get(config["permission_check_url"], timeout=10)
        if response.status_code != 200:
            raise XlwingsError(
                f"Failed to connect to permission server. Error {response.status_code}."
            )
        response = response.json()
        checked_files = []
        for file_name in file_names:
            for module in response["modules"]:
                if file_name == module["file_name"]:
                    correct_sha256 = file_hashes[file_name] == module["sha256"]
                    permitted_machine = (
                        module["machine_names"] == "*"
                        or "*" in module["machine_names"]
                        or socket.gethostname() in module["machine_names"]
                    )
                    if correct_sha256 and permitted_machine:
                        checked_files.append(file_name)
                        break
                    else:
                        raise XlwingsError(
                            f"Failed to get permission for the following file: "
                            f"{file_name}"
                        )
        missing_permissions = set(file_names).difference(set(checked_files))
        if missing_permissions:
            raise XlwingsError(
                f"Failed to get permission for the following file(s): "
                f"{', '.join(missing_permissions)}"
            )
    elif method == "POST":
        payload = {"machine_name": socket.gethostname(), "modules": []}
        for file_name, sha256 in file_hashes.items():
            payload["modules"].append({"file_name": file_name, "sha256": sha256})
        headers = (
            {"Authorization": config["permission_check_authorization"]}
            if config.get("permission_check_authorization")
            else None
        )
        response = requests.post(
            config["permission_check_url"],
            json=payload,
            timeout=10,
            headers=headers,
        )

        if response.status_code == 200:
            return True
        else:
            raise XlwingsError(
                f"Failed to get permission for the following file(s): "
                f"{', '.join(file_names)}. Error {response.status_code}."
            )
    else:
        raise ValueError("PERMISSION_CHECK_URL must be either GET or POST.")



import os
import sys
import unittest

import xlwings as xw

from .common import TestBase, this_dir

if sys.version_info[0] >= 3 and sys.version_info[1] >= 6:
    import pathlib
else:
    pathlib = None


class TestBooks(TestBase):
    def test_indexing(self):
        self.assertEqual(self.app1.books[0], self.app1.books(1))

    def test_len(self):
        count = self.app1.books.count
        wb = self.app1.books.add()
        self.assertEqual(len(self.app1.books), count + 1)

    def test_count(self):
        self.assertEqual(len(self.app1.books), self.app1.books.count)

    def test_add(self):
        current_count = self.app1.books.count
        self.app1.books.add()
        self.assertEqual(len(self.app1.books), current_count + 1)

    def test_open(self):
        fullname = os.path.join(this_dir, "test book.xlsx")
        wb = self.app1.books.open(fullname)
        self.assertEqual(self.app1.books.active, wb)

        wb2 = self.app1.books.open(fullname)  # Should not reopen
        self.assertEqual(wb, wb2)

    @unittest.skipIf(pathlib is None, "pathlib unavailable")
    def test_open_pathlib(self):
        fullname = pathlib.Path(this_dir) / "test book.xlsx"
        wb = self.app1.books.open(fullname)
        self.assertEqual(self.app1.books.active, wb)

        wb2 = self.app1.books.open(fullname)  # Should not reopen
        self.assertEqual(wb, wb2)

    def test_open_bad_name(self):
        fullname = os.path.join(this_dir, "no book.xlsx")
        with self.assertRaises(FileNotFoundError):
            self.app1.books.open(fullname)

    def test_iter(self):
        for ix, wb in enumerate(self.app1.books):
            self.assertEqual(self.app1.books[ix], wb)


class TestBook(TestBase):
    def test_instantiate_unsaved(self):
        self.wb1.sheets[0].range("B2").value = 123
        wb2 = self.app1.books[self.wb1.name]
        self.assertEqual(wb2.sheets[0].range("B2").value, 123)

    def test_instantiate_two_unsaved(self):
        """Covers GH Issue #63"""
        wb1 = self.wb1
        wb2 = self.app1.books.add()

        wb2.sheets[0].range("A1").value = 2.0
        wb1.sheets[0].range("A1").value = 1.0

        self.assertEqual(wb2.sheets[0].range("A1").value, 2.0)
        self.assertEqual(wb1.sheets[0].range("A1").value, 1.0)

    def test_instantiate_saved_by_name(self):
        wb1 = self.app1.books.open(
            os.path.join(os.path.dirname(os.path.abspath(__file__)), "test book.xlsx")
        )
        wb1.sheets[0].range("A1").value = "xx"
        wb2 = self.app1.books["test book.xlsx"]
        self.assertEqual(wb2.sheets[0].range("A1").value, "xx")

    def test_instantiate_saved_by_fullpath(self):
        # unicode name of book, but not unicode path
        wb = self.app1.books.add()
        if sys.platform.startswith("darwin") and self.app1.version.major >= 15:
            dst = os.path.join(
                os.path.expanduser("~")
                + "/Library/Containers/com.microsoft.Excel/Data/",
                "üni cöde.xlsx",
            )
        else:
            dst = os.path.join(
                os.path.dirname(os.path.abspath(__file__)), "üni cöde.xlsx"
            )
        if os.path.isfile(dst):
            os.remove(dst)
        wb.save(dst)
        wb2 = self.app1.books.open(dst)  # Book is open
        wb2.sheets[0].range("A1").value = 1
        wb2.save()
        wb2.close()
        wb3 = self.app1.books.open(dst)  # Book is closed
        self.assertEqual(wb3.sheets[0].range("A1").value, 1.0)
        wb3.close()
        os.remove(dst)

    def test_active(self):
        self.wb2.sheets[0].range("A1").value = "active_book"  # 2nd instance
        self.assertEqual(
            self.app2.books.active.sheets[0].range("A1").value, "active_book"
        )

    def test_mock_caller(self):
        path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "test book.xlsx"
        )

        wb = self.app1.books.open(path)
        wb.set_mock_caller()
        wb2 = xw.Book.caller()
        wb2.sheets[0].range("A1").value = 333
        self.assertEqual(wb2.sheets[0].range("A1").value, 333)

    def test_macro(self):
        # NOTE: Uncheck Macro security check in Excel
        _none = None if sys.platform.startswith("win") else ""

        src = os.path.abspath(os.path.join(this_dir, "macro book.xlsm"))
        wb = self.app1.books.open(src)

        test1 = wb.macro("Module1.Test1")
        test2 = wb.macro("Module1.Test2")
        test3 = wb.macro("Module1.Test3")
        test4 = wb.macro("Test4")
        test5 = wb.macro("Test5")

        res1 = test1("Test1a", "Test1b")

        self.wb1.sheets[0].range("A7").value = "range test"

        self.assertEqual(res1, 1)
        self.assertEqual(test2(), 2)
        self.assertEqual(test3("Test3a", "Test3b"), _none)
        self.assertEqual(test4(), _none)
        self.assertEqual(wb.sheets[0].range("A1").value, "Test1a")
        self.assertEqual(wb.sheets[0].range("A2").value, "Test1b")
        self.assertEqual(wb.sheets[0].range("A3").value, "Test2")
        self.assertEqual(wb.sheets[0].range("A4").value, "Test3a")
        self.assertEqual(wb.sheets[0].range("A5").value, "Test3b")
        self.assertEqual(wb.sheets[0].range("A6").value, "Test4")
        test5(self.wb1.sheets[0].range("A7"))
        self.assertEqual(wb.sheets[0].range("A8").value, "range test")
        test5(self.wb1.sheets[0].range("A7").api)
        self.assertEqual(wb.sheets[0].range("A8").value, "range test")

    def test_name(self):
        wb = self.app1.books.open(os.path.join(this_dir, "test book.xlsx"))
        self.assertEqual(wb.name, "test book.xlsx")

    @unittest.skipIf(pathlib is None, "pathlib unavailable")
    def test_name_pathlib(self):
        wb = self.app1.books.open(pathlib.Path(this_dir) / "test book.xlsx")
        self.assertEqual(wb.name, "test book.xlsx")

    def test_sheets(self):
        self.assertEqual(len(self.wb1.sheets), 3)

    def test_app(self):
        # Win Excel 2016 struggles with this test in any other more meaningful way
        self.assertEqual(self.app1.books[0].app.books[0], self.app1.books[0])

    def test_close(self):
        wb = self.app1.books.add()
        count = self.app1.books.count
        wb.close()
        self.assertEqual(len(self.app1.books), count - 1)

    def test_save_naked(self):
        if sys.platform.startswith("darwin") and self.app1.version.major >= 15:
            folder = (
                os.path.expanduser("~")
                + "/Library/Containers/com.microsoft.Excel/Data/"
            )
            if os.path.isdir(folder):
                os.chdir(folder)

        cwd = os.getcwd()
        target_file_path = os.path.join(cwd, self.wb1.name + ".xlsx")
        if os.path.isfile(target_file_path):
            os.remove(target_file_path)

        self.wb1.save()

        self.assertTrue(os.path.isfile(target_file_path))

        self.app1.books[os.path.basename(target_file_path)].close()
        if os.path.isfile(target_file_path):
            os.remove(target_file_path)

    def test_save_path(self):
        if sys.platform.startswith("darwin") and self.app1.version.major >= 15:
            folder = (
                os.path.expanduser("~")
                + "/Library/Containers/com.microsoft.Excel/Data/"
            )
            if os.path.isdir(folder):
                os.chdir(folder)

        cwd = os.getcwd()
        target_file_path = os.path.join(cwd, "TestFile.xlsx")
        if os.path.isfile(target_file_path):
            os.remove(target_file_path)

        self.wb1.save(target_file_path)

        self.assertTrue(os.path.isfile(target_file_path))

        self.app1.books[os.path.basename(target_file_path)].close()
        if os.path.isfile(target_file_path):
            os.remove(target_file_path)

    @unittest.skipIf(pathlib is None, "pathlib unavailable")
    def test_save_path_pathlib(self):
        if sys.platform.startswith("darwin") and self.app1.version.major >= 15:
            folder = (
                os.path.expanduser("~")
                + "/Library/Containers/com.microsoft.Excel/Data/"
            )
            if os.path.isdir(folder):
                os.chdir(folder)

        cwd = os.getcwd()
        target_file_path = pathlib.Path(cwd) / "TestFile.xlsx"
        if os.path.isfile(target_file_path):
            os.remove(target_file_path)

        self.wb1.save(target_file_path)

        self.assertTrue(os.path.isfile(target_file_path))

        self.app1.books[os.path.basename(target_file_path)].close()
        if os.path.isfile(target_file_path):
            os.remove(target_file_path)

    def test_fullname(self):
        fullname = os.path.join(this_dir, "test book.xlsx")
        wb = self.app1.books.open(fullname)
        self.assertEqual(wb.fullname.lower(), fullname.lower())

    @unittest.skipIf(pathlib is None, "pathlib unavailable")
    def test_fullname_pathlib(self):
        fullname = pathlib.Path(this_dir) / "test book.xlsx"
        wb = self.app1.books.open(fullname)
        self.assertEqual(wb.fullname.lower(), os.fspath(fullname).lower())

    def test_names(self):
        names = self.wb1.names
        self.assertEqual(len(names), 0)

    def test_activate(self):
        if sys.platform.startswith("win") and self.app1.version.major > 14:
            with self.assertRaises(Exception):
                wb1 = self.app1.books.add()
                wb2 = self.app2.books.add()
                wb1.activate()
        else:
            wb1 = self.app1.books.add()
            wb2 = self.app2.books.add()
            wb1.activate()
            self.assertEqual(xw.books.active, wb1)
            wb2.activate()
            self.assertEqual(xw.books.active, wb2)

    def test_selection(self):
        self.wb1.sheets[0].range("B10").select()
        self.assertEqual(self.wb1.selection.address, "$B$10")
        self.wb2.sheets[0].range("A2:C3").select()
        self.assertEqual(self.wb2.selection.address, "$A$2:$C$3")

    def test_sheet(self):
        self.wb1.sheets.add()
        self.assertEqual(len(self.wb1.sheets), 4)

    def test_sheets_names(self):
        self.assertEqual(self.wb1.sheet_names, ["Sheet1", "Sheet2", "Sheet3"])


if __name__ == "__main__":
    unittest.main()



from fastapi import Body, FastAPI, status
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import PlainTextResponse

import xlwings as xw

app = FastAPI()


@app.post("/hello")
def hello(data: dict = Body):
    # Instantiate a Book object with the deserialized request body
    with xw.Book(json=data) as book:
        # Use xlwings as usual
        sheet = book.sheets[0]
        cell = sheet["A1"]
        if cell.value == "Hello xlwings!":
            cell.value = "Bye xlwings!"
        else:
            cell.value = "Hello xlwings!"

        # Pass the following back as the response
        return book.json()


@app.exception_handler(Exception)
async def exception_handler(request, exception):
    # This handles all exceptions, so you may want to make this more restrictive
    return PlainTextResponse(
        str(exception), status_code=status.HTTP_500_INTERNAL_SERVER_ERROR
    )


# Office Scripts and custom functions in Excel on the web require CORS
cors_app = CORSMiddleware(
    app=app,
    allow_origins="*",
    allow_methods=["POST"],
    allow_headers=["*"],
    allow_credentials=True,
)


if __name__ == "__main__":
    import uvicorn

    uvicorn.run("main:cors_app", host="127.0.0.1", port=8000, reload=True)



"""
Required Notice: Copyright (C) Zoomer Analytics GmbH.

xlwings PRO is dual-licensed under one of the following licenses:

* PolyForm Noncommercial License 1.0.0 (for noncommercial use):
  https://polyformproject.org/licenses/noncommercial/1.0.0
* xlwings PRO License (for commercial use):
  https://github.com/xlwings/xlwings/blob/main/LICENSE_PRO.txt

Commercial licenses can be purchased at https://www.xlwings.org
"""

import inspect
from textwrap import dedent

from .. import XlwingsError, __version__, conversion


def func_sig(f):
    sig = inspect.signature(f)
    vararg = None
    args = []
    defaults = []
    for param in sig.parameters.values():
        if param.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD:
            args.append(param.name)
            if param.default is not inspect.Signature.empty:
                defaults.append(param.default)
        elif param.kind is inspect.Parameter.VAR_POSITIONAL:
            args.append(param.name)
            vararg = param.name
        else:
            raise XlwingsError("xlwings does not support UDFs with keyword arguments")
    return {"args": args, "defaults": defaults, "vararg": vararg}


def check_bool(kw, default, **func_kwargs):
    if kw in func_kwargs:
        check = func_kwargs.pop(kw)
        if isinstance(check, bool):
            return check
        raise XlwingsError(f'{kw} only takes boolean values. ("{check}" provided).')
    return default


def xlfunc(f=None, **kwargs):
    def inner(f):
        if not hasattr(f, "__xlfunc__"):
            xlf = f.__xlfunc__ = {}
            xlf["name"] = f.__name__
            xlargs = xlf["args"] = []
            xlargmap = xlf["argmap"] = {}
            sig = func_sig(f)
            num_args = len(sig["args"])
            num_defaults = len(sig["defaults"])
            num_required_args = num_args - num_defaults
            if sig["vararg"] and num_defaults > 0:
                raise XlwingsError(
                    "xlwings does not support UDFs "
                    "with both optional and variable length arguments"
                )
            for var_pos, var_name in enumerate(sig["args"]):
                arg_info = {
                    "name": var_name,
                    "pos": var_pos,
                    "doc": f"Positional argument {var_pos + 1}",
                    "vararg": var_name == sig["vararg"],
                    "options": {},
                }
                if var_pos >= num_required_args:
                    arg_info["optional"] = sig["defaults"][var_pos - num_required_args]
                xlargs.append(arg_info)
                xlargmap[var_name] = xlargs[-1]
            xlf["ret"] = {
                "doc": f.__doc__
                if f.__doc__ is not None
                else f"Python function '{f.__name__}'",
                "options": {},
            }
        f.__xlfunc__["volatile"] = check_bool("volatile", default=False, **kwargs)
        # If there's a global namespace defined in the manifest, this will be the
        # sub-namespace, i.e. NAMESPACE.SUBNAMESPACE.FUNCTIONNAME
        f.__xlfunc__["namespace"] = kwargs.get("namespace")
        f.__xlfunc__["help_url"] = kwargs.get("help_url")
        return f

    if f is None:
        return inner
    else:
        return inner(f)


def xlret(convert=None, **kwargs):
    if convert is not None:
        kwargs["convert"] = convert

    def inner(f):
        xlf = xlfunc(f).__xlfunc__
        xlr = xlf["ret"]
        xlr["options"].update(kwargs)
        return f

    return inner


def xlarg(arg, convert=None, **kwargs):
    if convert is not None:
        kwargs["convert"] = convert

    def inner(f):
        xlf = xlfunc(f).__xlfunc__
        if arg not in xlf["argmap"]:
            raise Exception(f"Invalid argument name '{arg}'.")
        xla = xlf["argmap"][arg]
        if "doc" in kwargs:
            xla["doc"] = kwargs.pop("doc")
        xla["options"].update(kwargs)
        return f

    return inner


def to_scalar(arg):
    if isinstance(arg, (list, tuple)) and len(arg) == 1:
        if isinstance(arg[0], (list, tuple)) and len(arg[0]) == 1:
            arg = arg[0][0]
        else:
            arg = arg[0]
    return arg


async def custom_functions_call(data, module):
    func_name = data["func_name"]
    args = data["args"]
    func = getattr(module, func_name)
    func_info = func.__xlfunc__
    args_info = func_info["args"]
    ret_info = func_info["ret"]

    if data["version"] != __version__:
        raise XlwingsError(
            "xlwings version mismatch: please restart Excel or "
            "right-click on the task pane and select 'reload'!"
        )

    args = list(args)
    # Turn varargs into regular arguments (remove the outermost list)
    for i, arg in enumerate(args):
        arg_info = args_info[min(i, len(args_info) - 1)]
        if arg_info["vararg"]:
            del args[i]
            args[i:i] = arg

    for i, arg in enumerate(args):
        arg_info = args_info[min(i, len(args_info) - 1)]
        arg = to_scalar(arg)
        if arg is None:
            args[i] = arg_info.get("optional", None)
        else:
            args[i] = conversion.read(
                None, arg, arg_info["options"], engine_name="officejs"
            )
    if inspect.iscoroutinefunction(func):
        ret = await func(*args)
    else:
        ret = func(*args)

    if "date_format" not in ret_info["options"]:
        ret_info["options"]["date_format"] = locale_to_shortdate[
            data["content_language"].lower()
        ]
    ret_info["options"]["runtime"] = data["runtime"]
    ret = conversion.write(ret, None, ret_info["options"], engine_name="officejs")
    return ret


def custom_functions_code(
    module, custom_functions_call_path="/xlwings/custom-functions-call"
):
    js = """\
         async function base() {
           // Turn arguments into an array, the last one is the invocation parameter
           let argsArr = Array.prototype.slice.call(arguments);
           let func_name = argsArr[0];
           let args = argsArr.slice(1, -1);
           let invocation = argsArr[argsArr.length - 1];
           // headers
           let headers = {};
           headers["Content-Type"] = "application/json";
           headers["Authorization"] = await globalThis.getAuth();
           let runtime;
           if (
             Office.context.requirements.isSetSupported("CustomFunctionsRuntime", "1.4")
           ) {
             runtime = "1.4";
           } else if (
             Office.context.requirements.isSetSupported("CustomFunctionsRuntime", "1.3")
           ) {
             runtime = "1.3";
           } else if (
             Office.context.requirements.isSetSupported("CustomFunctionsRuntime", "1.2")
           ) {
             runtime = "1.2";
           } else {
             runtime = "1.1";
           }
           let response = await fetch(
             window.location.origin + "custom_functions_call_path",
             {
               method: "POST",
               headers: headers,
               body: JSON.stringify({
                 func_name: func_name,
                 args: args,
                 caller_address: invocation.address,
                 formula_name: invocation.functionName,
                 content_language: Office.context.contentLanguage,
                 version: "xlwings_version",
                 runtime: runtime,
               }),
             }
           );
           if (response.status !== 200) {
             let errMsg = await response.text();
             // Error message only visible by hovering over the error flag!
             if (
               Office.context.requirements.isSetSupported(
                 "CustomFunctionsRuntime",
                 "1.2"
               )
             ) {
               let error = new CustomFunctions.Error(
                 CustomFunctions.ErrorCode.invalidValue,
                 errMsg
               );
               throw error;
             } else {
               return [[errMsg]];
             }
           } else {
             rawData = await response.json();
           }
           return rawData.result;
         }
    """.replace(
        "xlwings_version", __version__
    ).replace(
        "custom_functions_call_path", custom_functions_call_path
    )  # format string would require to double all curly braces
    js = dedent(js)
    for name, obj in inspect.getmembers(module):
        if hasattr(obj, "__xlfunc__"):
            xlfunc = obj.__xlfunc__
            func_name = xlfunc["name"]
            js += dedent(
                f"""\
            async function {func_name}() {{
                args = ["{func_name}"]
                args.push.apply(args, arguments);
                return await base.apply(null, args);
            }}
            CustomFunctions.associate("{func_name.upper()}", {func_name});
            """
            )
    return js


def custom_functions_meta(module):
    funcs = []
    for name, obj in inspect.getmembers(module):
        if hasattr(obj, "__xlfunc__"):
            xlfunc = obj.__xlfunc__
            func = {}
            func["description"] = xlfunc["ret"]["doc"]
            if xlfunc["help_url"]:
                func["helpUrl"] = xlfunc["help_url"]
            func["id"] = xlfunc["name"].upper()
            if xlfunc["namespace"]:
                func["name"] = f"{xlfunc['namespace'].upper()}.{xlfunc['name'].upper()}"
            else:
                func["name"] = xlfunc["name"].upper()
            func["options"] = {
                "requiresAddress": True,
                "requiresParameterAddresses": True,
            }
            if xlfunc["volatile"]:
                func["options"]["volatile"] = True
            func["result"] = {"dimensionality": "matrix", "type": "any"}

            params = []
            for arg in xlfunc["args"]:
                param = {}
                param["description"] = arg["doc"]
                param["name"] = arg["name"]
                param["dimensionality"] = "matrix"
                param["type"] = "any"
                if "optional" in arg:
                    param["optional"] = True
                elif arg["vararg"]:
                    param["repeating"] = True
                params.append(param)
            func["parameters"] = params
            funcs.append(func)
    return {
        "allowCustomDataForDataTypeAny": True,
        "allowErrorForDataTypeAny": True,
        "functions": funcs,
    }


locale_to_shortdate = {
    # This is using the locales from https://github.com/OfficeDev/office-js/tree/release/dist
    # matched with values from https://stackoverflow.com/a/9893752/918626
    "af-za": "yyyy/mm/dd",
    "am-et": "d/m/yyyy",
    "ar-ae": "dd/mm/yyyy",
    "ar-bh": "dd/mm/yyyy",
    "ar-dz": "dd-mm-yyyy",
    "ar-eg": "dd/mm/yyyy",
    "ar-iq": "dd/mm/yyyy",
    "ar-jo": "dd/mm/yyyy",
    "ar-kw": "dd/mm/yyyy",
    "ar-lb": "dd/mm/yyyy",
    "ar-ly": "dd/mm/yyyy",
    "ar-ma": "dd-mm-yyyy",
    "ar-om": "dd/mm/yyyy",
    "ar-qa": "dd/mm/yyyy",
    "ar-sa": "dd/mm/yy",
    "ar-sy": "dd/mm/yyyy",
    "ar-tn": "dd-mm-yyyy",
    "ar-ye": "dd/mm/yyyy",
    "az-latn-az": "dd.mm.yyyy",
    "be-by": "dd.mm.yyyy",
    "bg-bg": "dd.m.yyyy",
    "bn-in": "dd-mm-yy",
    "bs-latn-ba": "d.m.yyyy",
    "ca-es": "dd/mm/yyyy",
    "cs-cz": "d.m.yyyy",
    "cy-gb": "dd/mm/yyyy",
    "da-dk": "dd-mm-yyyy",
    "de-at": "dd.mm.yyyy",
    "de-ch": "dd.mm.yyyy",
    "de-de": "dd.mm.yyyy",
    "de-li": "dd.mm.yyyy",
    "de-lu": "dd.mm.yyyy",
    "el-gr": "d/m/yyyy",
    "en-029": "mm/dd/yyyy",
    "en-au": "d/mm/yyyy",
    "en-bz": "dd/mm/yyyy",
    "en-ca": "dd/mm/yyyy",
    "en-gb": "dd/mm/yyyy",
    "en-ie": "dd/mm/yyyy",
    "en-in": "dd-mm-yyyy",
    "en-jm": "dd/mm/yyyy",
    "en-my": "d/m/yyyy",
    "en-nz": "d/mm/yyyy",
    "en-ph": "m/d/yyyy",
    "en-sg": "d/m/yyyy",
    "en-tt": "dd/mm/yyyy",
    "en-us": "m/d/yyyy",
    "en-za": "yyyy/mm/dd",
    "en-zw": "m/d/yyyy",
    "es-ar": "dd/mm/yyyy",
    "es-bo": "dd/mm/yyyy",
    "es-cl": "dd-mm-yyyy",
    "es-co": "dd/mm/yyyy",
    "es-cr": "dd/mm/yyyy",
    "es-do": "dd/mm/yyyy",
    "es-ec": "dd/mm/yyyy",
    "es-es": "dd/mm/yyyy",
    "es-gt": "dd/mm/yyyy",
    "es-hn": "dd/mm/yyyy",
    "es-mx": "dd/mm/yyyy",
    "es-ni": "dd/mm/yyyy",
    "es-pa": "mm/dd/yyyy",
    "es-pe": "dd/mm/yyyy",
    "es-pr": "dd/mm/yyyy",
    "es-py": "dd/mm/yyyy",
    "es-sv": "dd/mm/yyyy",
    "es-us": "m/d/yyyy",
    "es-uy": "dd/mm/yyyy",
    "es-ve": "dd/mm/yyyy",
    "et-ee": "d.mm.yyyy",
    "eu-es": "yyyy/mm/dd",
    "fa-ir": "mm/dd/yyyy",
    "fi-fi": "d.m.yyyy",
    "fil-ph": "m/d/yyyy",
    "fr-be": "d/mm/yyyy",
    "fr-ca": "yyyy-mm-dd",
    "fr-ch": "dd.mm.yyyy",
    "fr-fr": "dd/mm/yyyy",
    "fr-lu": "dd/mm/yyyy",
    "fr-mc": "dd/mm/yyyy",
    "ga-ie": "dd/mm/yyyy",
    "gl-es": "dd/mm/yy",
    "gu-in": "dd-mm-yy",
    "he-il": "dd/mm/yyyy",
    "hi-in": "dd-mm-yyyy",
    "hr-ba": "d.m.yyyy.",
    "hr-hr": "d.m.yyyy",
    "hu-hu": "yyyy. mm. dd.",
    "hy-am": "dd.mm.yyyy",
    "id-id": "dd/mm/yyyy",
    "is-is": "d.m.yyyy",
    "it-ch": "dd.mm.yyyy",
    "it-it": "dd/mm/yyyy",
    "ja-jp": "yyyy/mm/dd",
    "ka-ge": "dd.mm.yyyy",
    "kk-kz": "dd.mm.yyyy",
    "km-kh": "yyyy-mm-dd",
    "kn-in": "dd-mm-yy",
    "ko-kr": "yyyy. mm. dd",
    "lb-lu": "dd/mm/yyyy",
    "lo-la": "dd/mm/yyyy",
    "lt-lt": "yyyy.mm.dd",
    "lv-lv": "yyyy.mm.dd.",
    "mk-mk": "dd.mm.yyyy",
    "ml-in": "dd-mm-yy",
    "mn-mn": "yy.mm.dd",
    "mr-in": "dd-mm-yyyy",
    "ms-bn": "dd/mm/yyyy",
    "ms-my": "dd/mm/yyyy",
    "mt-mt": "dd/mm/yyyy",
    "nb-no": "dd.mm.yyyy",
    "ne-np": "m/d/yyyy",
    "nl-be": "d/mm/yyyy",
    "nl-nl": "d-m-yyyy",
    "nn-no": "dd.mm.yyyy",
    "pl-pl": "dd.mm.yyyy",
    "pt-br": "d/m/yyyy",
    "pt-pt": "dd-mm-yyyy",
    "ro-ro": "dd.mm.yyyy",
    "ru-ru": "dd.mm.yyyy",
    "si-lk": "yyyy-mm-dd",
    "sk-sk": "d. m. yyyy",
    "sl-si": "d.m.yyyy",
    "sq-al": "yyyy-mm-dd",
    "sr-cyrl-cs": "d.m.yyyy",
    "sr-cyrl-rs": "d.m.yyyy",
    "sr-latn-cs": "d.m.yyyy",
    "sr-latn-rs": "d.m.yyyy",
    "sv-fi": "d.m.yyyy",
    "sv-se": "yyyy-mm-dd",
    "sw-ke": "m/d/yyyy",
    "ta-in": "dd-mm-yyyy",
    "te-in": "dd-mm-yy",
    "th-th": "d/m/yyyy",
    "tr-tr": "dd.mm.yyyy",
    "uk-ua": "dd.mm.yyyy",
    "ur-pk": "dd/mm/yyyy",
    "vi-vn": "dd/mm/yyyy",
    "zh-cn": "yyyy/m/d",
    "zh-hk": "d/m/yyyy",
    "zh-mo": "d/m/yyyy",
    "zh-sg": "d/m/yyyy",
    "zh-tw": "yyyy/m/d",
}



import datetime as dt
import sys
import unittest
from collections import OrderedDict

import pytz

from .common import TestBase

# Optional dependencies
try:
    import numpy as np
    from numpy.testing import assert_array_equal
except ImportError:
    np = None
try:
    import pandas as pd
    from pandas import DataFrame
    from pandas.testing import assert_frame_equal, assert_series_equal
except ImportError:
    pd = None


def table_formatter(rng, values):
    """This is the formatter function"""
    # Header
    rng[0, :].color = "#A9D08E"

    # Rows
    for ix, row in enumerate(rng.rows[1:]):
        if ix % 2 == 0:
            row.color = "#D0CECE"  # Even rows


class TestConverter(TestBase):
    def test_transpose(self):
        self.wb1.sheets[0].range("A1").options(transpose=True).value = [
            [1.0, 2.0],
            [3.0, 4.0],
        ]
        self.assertEqual(
            self.wb1.sheets[0].range("A1:B2").value, [[1.0, 3.0], [2.0, 4.0]]
        )

    def test_dictionary(self):
        d = {"a": 1.0, "b": 2.0}
        self.wb1.sheets[0].range("A1").value = d
        self.assertEqual(d, self.wb1.sheets[0].range("A1:B2").options(dict).value)

    def test_ordereddictionary(self):
        d = OrderedDict({"a": 1.0, "b": 2.0})
        self.wb1.sheets[0].range("A1").value = d
        self.assertEqual(
            d, self.wb1.sheets[0].range("A1:B2").options(OrderedDict).value
        )

    def test_integers(self):
        """test_integers: Covers GH 227"""
        self.wb1.sheets[0].range("A99").value = 2147483647  # max SInt32
        self.assertEqual(self.wb1.sheets[0].range("A99").value, 2147483647)

        self.wb1.sheets[0].range("A100").value = 2147483648  # SInt32 < x < SInt64
        self.assertEqual(self.wb1.sheets[0].range("A100").value, 2147483648)

        self.wb1.sheets[0].range("A101").value = 10000000000000000000  # long
        self.assertEqual(self.wb1.sheets[0].range("A101").value, 10000000000000000000)

    def test_datetime_timezone(self):
        eastern = pytz.timezone("US/Eastern")
        dt_naive = dt.datetime(2002, 10, 27, 6, 0, 0)
        dt_tz = eastern.localize(dt_naive)
        self.wb1.sheets[0].range("F34").value = dt_tz
        self.assertEqual(self.wb1.sheets[0].range("F34").value, dt_naive)

    def test_date(self):
        date_1 = dt.date(2000, 12, 3)
        self.wb1.sheets[0].range("X1").value = date_1
        date_2 = self.wb1.sheets[0].range("X1").value
        self.assertEqual(date_1, dt.date(date_2.year, date_2.month, date_2.day))

    def test_list(self):
        # 1d List Row
        list_row_1d = [1.1, None, 3.3]
        self.wb1.sheets[0].range("A27").value = list_row_1d
        cells = self.wb1.sheets[0].range("A27:C27").value
        self.assertEqual(list_row_1d, cells)

        # 2d List Row
        list_row_2d = [[1.1, None, 3.3]]
        self.wb1.sheets[0].range("A29").value = list_row_2d
        cells = self.wb1.sheets[0].range("A29:C29").options(ndim=2).value
        self.assertEqual(list_row_2d, cells)

        # 1d List Col
        list_col = [[1.1], [None], [3.3]]
        self.wb1.sheets[0].range("A31").value = list_col
        cells = self.wb1.sheets[0].range("A31:A33").value
        self.assertEqual([i[0] for i in list_col], cells)
        # 2d List Col
        cells = self.wb1.sheets[0].range("A31:A33").options(ndim=2).value
        self.assertEqual(list_col, cells)

    def test_none(self):
        """test_none: Covers GH Issue #16"""
        # None
        self.wb1.sheets[0].range("A7").value = None
        self.assertEqual(None, self.wb1.sheets[0].range("A7").value)
        # List
        self.wb1.sheets[0].range("A7").value = [None, None]
        self.assertEqual(
            None, self.wb1.sheets[0].range("A7").expand("horizontal").value
        )

    def test_ndim2_scalar(self):
        """test_atleast_2d_scalar: Covers GH Issue #53a"""
        self.wb1.sheets[0].range("A50").value = 23
        result = self.wb1.sheets[0].range("A50").options(ndim=2).value
        self.assertEqual([[23]], result)

    def test_write_single_value_to_multicell_range(self):
        self.wb1.sheets[0].range("A1:B2").value = 5
        self.assertEqual(
            self.wb1.sheets[0].range("A1:B2").value, [[5.0, 5.0], [5.0, 5.0]]
        )

    def test_formatter(self):
        self.wb1.sheets[0]["A50"].options(formatter=table_formatter).value = [
            ["one", "two"],
            [1, 2],
            [3, 4],
            [5, 6],
        ]
        self.assertEqual(self.wb1.sheets[0]["A50:B50"].color, (169, 208, 142))
        self.assertEqual(self.wb1.sheets[0]["A51:B51"].color, (208, 206, 206))
        self.assertIsNone(self.wb1.sheets[0]["A52:B52"].color)


@unittest.skipIf(np is None, "numpy missing")
class TestNumpy(TestBase):
    def test_array(self):
        # 1d array
        array_1d = np.array([1.1, 2.2, np.nan, -4.4])
        self.wb1.sheets[0].range("A1").value = array_1d
        cells = self.wb1.sheets[0].range("A1:D1").options(np.array).value
        assert_array_equal(cells, array_1d)

        # 2d array
        array_2d = np.array([[1.1, 2.2, 3.3], [-4.4, 5.5, np.nan]])
        self.wb1.sheets[0].range("A4").value = array_2d
        cells = self.wb1.sheets[0].range("A4").options(np.array, expand="table").value
        assert_array_equal(cells, array_2d)

        # 1d array (ndim=2)
        self.wb1.sheets[0].range("A10").value = array_1d
        cells = self.wb1.sheets[0].range("A10:D10").options(np.array, ndim=2).value
        assert_array_equal(cells, np.atleast_2d(array_1d))

        # 2d array (ndim=2)
        self.wb1.sheets[0].range("A12").value = array_2d
        cells = (
            self.wb1.sheets[0]
            .range("A12")
            .options(np.array, ndim=2, expand="table")
            .value
        )
        assert_array_equal(cells, array_2d)

    def test_numpy_datetime(self):
        self.wb1.sheets[0].range("A55").value = np.datetime64("2005-02-25T03:30Z")
        self.assertEqual(
            self.wb1.sheets[0].range("A55").value, dt.datetime(2005, 2, 25, 3, 30)
        )

    def test_scalar_nan(self):
        """test_scalar_nan: Covers GH Issue #15"""
        self.wb1.sheets[0].range("A20").value = np.nan
        self.assertEqual(None, self.wb1.sheets[0].range("A20").value)

    def test_scalar_nan_float16(self):
        """test_scalar_nan: Covers GH Issue #1116"""
        self.wb1.sheets[0].range("A20").value = np.float16(np.nan)
        self.assertEqual(None, self.wb1.sheets[0].range("A20").value)

    def test_scalar_nan_float32(self):
        """test_scalar_nan: Covers GH Issue #1116"""
        self.wb1.sheets[0].range("A20").value = np.float32(np.nan)
        self.assertEqual(None, self.wb1.sheets[0].range("A20").value)

    def test_scalar_nan_float64(self):
        """test_scalar_nan: Covers GH Issue #1116"""
        self.wb1.sheets[0].range("A20").value = np.float64(np.nan)
        self.assertEqual(None, self.wb1.sheets[0].range("A20").value)

    def test_ndim2_scalar_as_array(self):
        """test_atleast_2d_scalar_as_array: Covers GH Issue #53b"""
        self.wb1.sheets[0].range("A50").value = 23
        result = self.wb1.sheets[0].range("A50").options(np.array, ndim=2).value
        self.assertEqual(np.array([[23]]), result)

    def test_float64(self):
        self.wb1.sheets[0].range("A1").value = np.float64(2)
        self.assertEqual(self.wb1.sheets[0].range("A1").value, 2.0)

    def test_int64(self):
        self.wb1.sheets[0].range("A1").value = np.int64(2)
        self.assertEqual(self.wb1.sheets[0].range("A1").value, 2.0)


@unittest.skipIf(pd is None, "pandas missing")
class TestPandas(TestBase):
    def test_dataframe_1(self):
        df_expected = pd.DataFrame(
            [[1, "test1"], [2, "test2"], [np.nan, None], [3.3, "test3"]],
            columns=["a", "b"],
        )
        self.wb1.sheets[0].range("A1").value = df_expected
        df_result = self.wb1.sheets[0].range("A1:C5").options(pd.DataFrame).value
        df_result.index = pd.Int64Index(df_result.index)
        assert_frame_equal(df_expected, df_result)

    def test_dataframe_2(self):
        """test_dataframe_2: Covers GH Issue #31"""
        df_expected = pd.DataFrame([1, 3, 5, np.nan, 6, 8], columns=["col1"])
        self.wb1.sheets[0].range("A9").value = df_expected
        cells = self.wb1.sheets[0].range("B9:B15").value
        df_result = DataFrame(cells[1:], columns=[cells[0]])
        assert_frame_equal(df_expected, df_result)

    def test_dataframe_multiindex(self):
        tuples = list(
            zip(
                *[
                    ["bar", "bar", "baz", "baz", "foo", "foo", "qux", "qux"],
                    ["one", "two", "one", "two", "one", "two", "one", "two"],
                    ["x", "x", "x", "x", "y", "y", "y", "y"],
                ]
            )
        )

        index = pd.MultiIndex.from_tuples(tuples, names=["first", "second", "third"])
        df_expected = pd.DataFrame(
            [
                [1.1, 2.2],
                [3.3, 4.4],
                [5.5, 6.6],
                [7.7, 8.8],
                [9.9, 10.10],
                [11.11, 12.12],
                [13.13, 14.14],
                [15.15, 16.16],
            ],
            index=index,
            columns=["one", "two"],
        )
        self.wb1.sheets[0].range("A20").value = df_expected
        cells = self.wb1.sheets[0].range("D20").expand("table").value
        multiindex = self.wb1.sheets[0].range("A20:C28").value
        ix = pd.MultiIndex.from_tuples(multiindex[1:], names=multiindex[0])
        df_result = DataFrame(cells[1:], columns=cells[0], index=ix)
        assert_frame_equal(df_expected, df_result)

    def test_dataframe_multiheader(self):
        header = [["Foo", "Foo", "Bar", "Bar", "Baz"], ["A", "B", "C", "D", "E"]]
        df_expected = pd.DataFrame(
            [
                [0.0, 1.0, 2.0, 3.0, 4.0],
                [0.0, 1.0, 2.0, 3.0, 4.0],
                [0.0, 1.0, 2.0, 3.0, 4.0],
                [0.0, 1.0, 2.0, 3.0, 4.0],
                [0.0, 1.0, 2.0, 3.0, 4.0],
                [0.0, 1.0, 2.0, 3.0, 4.0],
            ],
            columns=pd.MultiIndex.from_arrays(header),
        )
        self.wb1.sheets[0].range("A52").value = df_expected
        cells = self.wb1.sheets[0].range("B52").expand("table").value
        df_result = DataFrame(cells[2:], columns=pd.MultiIndex.from_arrays(cells[:2]))
        assert_frame_equal(df_expected, df_result)

    def test_dataframe_dateindex(self):
        rng = pd.date_range("1/1/2012", periods=10, freq="D")
        df_expected = pd.DataFrame(
            np.arange(50).reshape(10, 5) + 0.1,
            index=rng,
            columns=["one", "two", "three", "four", "five"],
        )
        self.wb1.sheets[0].range("A100").value = df_expected
        if sys.platform.startswith("win") and self.wb1.app.version == "14.0":
            self.wb1.sheets[0].range("A100").expand(
                "vertical"
            ).number_format = "dd/mm/yyyy"  # Hack for Excel 2010 bug, see GH #43
        cells = self.wb1.sheets[0].range("B100").expand("table").value
        index = self.wb1.sheets[0].range("A101").expand("vertical").value
        df_result = DataFrame(cells[1:], index=index, columns=cells[0])
        assert_frame_equal(df_expected, df_result)

    def test_read_df_0header_0index(self):
        self.wb1.sheets[0].range("A1").value = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

        df1 = pd.DataFrame([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]])

        df2 = (
            self.wb1.sheets[0]
            .range("A1:C3")
            .options(pd.DataFrame, header=0, index=False)
            .value
        )
        assert_frame_equal(df1, df2)

    def test_df_1header_0index(self):
        self.wb1.sheets[0].range("A1").options(
            pd.DataFrame, index=False, header=True
        ).value = pd.DataFrame([[1.0, 2.0], [3.0, 4.0]], columns=["a", "b"])
        df = (
            self.wb1.sheets[0]
            .range("A1")
            .options(pd.DataFrame, index=False, header=True, expand="table")
            .value
        )
        assert_frame_equal(
            df, pd.DataFrame([[1.0, 2.0], [3.0, 4.0]], columns=["a", "b"])
        )

    def test_df_0header_1index(self):
        self.wb1.sheets[0].range("A1").options(
            pd.DataFrame, index=True, header=False
        ).value = pd.DataFrame([[1.0, 2.0], [3.0, 4.0]], index=[10.0, 20.0])
        df = (
            self.wb1.sheets[0]
            .range("A1")
            .options(pd.DataFrame, index=True, header=False, expand="table")
            .value
        )
        assert_frame_equal(
            df, pd.DataFrame([[1.0, 2.0], [3.0, 4.0]], index=[10.0, 20.0])
        )

    def test_read_df_1header_1namedindex(self):
        self.wb1.sheets[0].range("A1").value = [
            ["ix1", "c", "d", "c"],
            [1, 1, 2, 3],
            [2, 4, 5, 6],
        ]

        df1 = pd.DataFrame(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]],
            index=[1.0, 2.0],
            columns=["c", "d", "c"],
        )
        df1.index.name = "ix1"

        df2 = self.wb1.sheets[0].range("A1:D3").options(pd.DataFrame).value
        assert_frame_equal(df1, df2)

    def test_read_df_1header_1unnamedindex(self):
        self.wb1.sheets[0].range("A1").value = [
            [None, "c", "d", "c"],
            [1, 1, 2, 3],
            [2, 4, 5, 6],
        ]

        df1 = pd.DataFrame(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]],
            index=pd.Index([1.0, 2.0]),
            columns=["c", "d", "c"],
        )

        df2 = self.wb1.sheets[0].range("A1:D3").options(pd.DataFrame).value

        assert_frame_equal(df1, df2)

    def test_read_df_2header_1namedindex(self):
        self.wb1.sheets[0].range("A1").value = [
            [None, "a", "a", "b"],
            ["ix1", "c", "d", "c"],
            [1, 1, 2, 3],
            [2, 4, 5, 6],
        ]

        df1 = pd.DataFrame(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]],
            index=[1.0, 2.0],
            columns=pd.MultiIndex.from_arrays([["a", "a", "b"], ["c", "d", "c"]]),
        )
        df1.index.name = "ix1"

        df2 = self.wb1.sheets[0].range("A1:D4").options(pd.DataFrame, header=2).value

        assert_frame_equal(df1, df2)

    def test_read_df_2header_1unnamedindex(self):
        self.wb1.sheets[0].range("A1").value = [
            [None, "a", "a", "b"],
            [None, "c", "d", "c"],
            [1, 1, 2, 3],
            [2, 4, 5, 6],
        ]

        df1 = pd.DataFrame(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]],
            index=pd.Index([1, 2]),
            columns=pd.MultiIndex.from_arrays([["a", "a", "b"], ["c", "d", "c"]]),
        )

        df2 = self.wb1.sheets[0].range("A1:D4").options(pd.DataFrame, header=2).value
        df2.index = pd.Int64Index(df2.index)

        assert_frame_equal(df1, df2)

    def test_read_df_2header_2namedindex(self):
        self.wb1.sheets[0].range("A1").value = [
            [None, None, "a", "a", "b"],
            ["x1", "x2", "c", "d", "c"],
            ["a", 1, 1, 2, 3],
            ["a", 2, 4, 5, 6],
            ["b", 1, 7, 8, 9],
        ]

        df1 = pd.DataFrame(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]],
            index=pd.MultiIndex.from_arrays(
                [["a", "a", "b"], [1.0, 2.0, 1.0]], names=["x1", "x2"]
            ),
            columns=pd.MultiIndex.from_arrays([["a", "a", "b"], ["c", "d", "c"]]),
        )

        df2 = (
            self.wb1.sheets[0]
            .range("A1:E5")
            .options(pd.DataFrame, header=2, index=2)
            .value
        )
        assert_frame_equal(df1, df2)

    def test_read_df_2header_2unnamedindex(self):
        self.wb1.sheets[0].range("A1").value = [
            [None, None, "a", "a", "b"],
            [None, None, "c", "d", "c"],
            ["a", 1, 1, 2, 3],
            ["a", 2, 4, 5, 6],
            ["b", 1, 7, 8, 9],
        ]

        df1 = pd.DataFrame(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]],
            index=pd.MultiIndex.from_arrays([["a", "a", "b"], [1.0, 2.0, 1.0]]),
            columns=pd.MultiIndex.from_arrays([["a", "a", "b"], ["c", "d", "c"]]),
        )

        df2 = (
            self.wb1.sheets[0]
            .range("A1:E5")
            .options(pd.DataFrame, header=2, index=2)
            .value
        )
        assert_frame_equal(df1, df2)

    def test_read_df_1header_2namedindex(self):
        self.wb1.sheets[0].range("A1").value = [
            ["x1", "x2", "a", "d", "c"],
            ["a", 1, 1, 2, 3],
            ["a", 2, 4, 5, 6],
            ["b", 1, 7, 8, 9],
        ]

        df1 = pd.DataFrame(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]],
            index=pd.MultiIndex.from_arrays(
                [["a", "a", "b"], [1.0, 2.0, 1.0]], names=["x1", "x2"]
            ),
            columns=["a", "d", "c"],
        )

        df2 = (
            self.wb1.sheets[0]
            .range("A1:E4")
            .options(pd.DataFrame, header=1, index=2)
            .value
        )
        assert_frame_equal(df1, df2)

    def test_timeseries_1(self):
        rng = pd.date_range("1/1/2012", periods=10, freq="D")
        series_expected = pd.Series(np.arange(len(rng)) + 0.1, rng)
        self.wb1.sheets[0].range("A40").options(header=False).value = series_expected
        if sys.platform.startswith("win") and self.wb1.app.version == "14.0":
            self.wb1.sheets[0].range("A40").expand(
                "vertical"
            ).number_format = "dd/mm/yyyy"  # Hack for Excel 2010 bug, see GH #43
        series_result = (
            self.wb1.sheets[0].range("A40:B49").options(pd.Series, header=False).value
        )
        assert_series_equal(series_expected, series_result)

    def test_read_series_noheader_noindex(self):
        self.wb1.sheets[0].range("A1").value = [[1.0], [2.0], [3.0]]
        s = (
            self.wb1.sheets[0]
            .range("A1:A3")
            .options(pd.Series, index=False, header=False)
            .value
        )
        assert_series_equal(s, pd.Series([1.0, 2.0, 3.0]))

    def test_read_series_noheader_index(self):
        self.wb1.sheets[0].range("A1").value = [[10.0, 1.0], [20.0, 2.0], [30.0, 3.0]]
        s = (
            self.wb1.sheets[0]
            .range("A1:B3")
            .options(pd.Series, index=True, header=False)
            .value
        )
        assert_series_equal(s, pd.Series([1.0, 2.0, 3.0], index=[10.0, 20.0, 30.0]))

    def test_read_series_header_noindex(self):
        self.wb1.sheets[0].range("A1").value = [["name"], [1.0], [2.0], [3.0]]
        s = (
            self.wb1.sheets[0]
            .range("A1:A4")
            .options(pd.Series, index=False, header=True)
            .value
        )
        assert_series_equal(s, pd.Series([1.0, 2.0, 3.0], name="name"))

    def test_read_series_header_index(self):
        # Named index
        self.wb1.sheets[0].range("A1").value = [
            ["ix", "name"],
            [10.0, 1.0],
            [20.0, 2.0],
            [30.0, 3.0],
        ]
        s = (
            self.wb1.sheets[0]
            .range("A1:B4")
            .options(pd.Series, index=True, header=True)
            .value
        )
        assert_series_equal(
            s,
            pd.Series(
                [1.0, 2.0, 3.0],
                name="name",
                index=pd.Index([10.0, 20.0, 30.0], name="ix"),
            ),
        )

        # Nameless index
        self.wb1.sheets[0].range("A1").value = [
            [None, "name"],
            [10.0, 1.0],
            [20.0, 2.0],
            [30.0, 3.0],
        ]
        s = (
            self.wb1.sheets[0]
            .range("A1:B4")
            .options(pd.Series, index=True, header=True)
            .value
        )
        assert_series_equal(
            s, pd.Series([1.0, 2.0, 3.0], name="name", index=[10.0, 20.0, 30.0])
        )

    def test_write_series_noheader_noindex(self):
        self.wb1.sheets[0].range("A1").options(index=False).value = pd.Series(
            [1.0, 2.0, 3.0]
        )
        self.assertEqual(
            [[1.0], [2.0], [3.0]],
            self.wb1.sheets[0].range("A1").options(ndim=2, expand="table").value,
        )

    def test_write_series_noheader_index(self):
        self.wb1.sheets[0].range("A1").options(index=True).value = pd.Series(
            [1.0, 2.0, 3.0], index=[10.0, 20.0, 30.0]
        )
        self.assertEqual(
            [[10.0, 1.0], [20.0, 2.0], [30.0, 3.0]],
            self.wb1.sheets[0].range("A1").options(ndim=2, expand="table").value,
        )

    def test_write_series_header_noindex(self):
        self.wb1.sheets[0].range("A1").options(index=False).value = pd.Series(
            [1.0, 2.0, 3.0], name="name"
        )
        self.assertEqual(
            [["name"], [1.0], [2.0], [3.0]],
            self.wb1.sheets[0].range("A1").options(ndim=2, expand="table").value,
        )

    def test_write_series_header_index(self):
        # Named index
        self.wb1.sheets[0].range("A1").value = pd.Series(
            [1.0, 2.0, 3.0], name="name", index=pd.Index([10.0, 20.0, 30.0], name="ix")
        )
        self.assertEqual(
            [["ix", "name"], [10.0, 1.0], [20.0, 2.0], [30.0, 3.0]],
            self.wb1.sheets[0].range("A1").options(ndim=2, expand="table").value,
        )

        # Nameless index
        self.wb1.sheets[0].range("A1").value = pd.Series(
            [1.0, 2.0, 3.0], name="name", index=[10.0, 20.0, 30.0]
        )
        self.assertEqual(
            [[None, "name"], [10.0, 1.0], [20.0, 2.0], [30.0, 3.0]],
            self.wb1.sheets[0].range("A1:B4").options(ndim=2).value,
        )

    def test_dataframe_timezone(self):
        np_dt = np.datetime64(1434149887000, "ms")
        ix = pd.DatetimeIndex(data=[np_dt], tz="GMT")
        df = pd.DataFrame(data=[1], index=ix, columns=["A"])
        self.wb1.sheets[0].range("A1").value = df
        self.assertEqual(
            self.wb1.sheets[0].range("A2").value, dt.datetime(2015, 6, 12, 22, 58, 7)
        )

    def test_NaT(self):
        df = pd.DataFrame(
            [pd.Timestamp("20120102"), np.nan], index=[0.0, 1.0], columns=["one"]
        )
        self.wb1.sheets[0].range("A1").value = df
        assert_frame_equal(
            df,
            self.wb1.sheets[0].range("A1").options(pd.DataFrame, expand="table").value,
        )

    def test_period_index(self):
        idx = pd.PeriodIndex(year=[2000, 2002], quarter=[1, 3])
        df = pd.DataFrame(index=idx, data=[1, 2])
        self.wb1.sheets[0].range("A1").value = df
        self.assertEqual(
            [[None, 0.0], ["2000Q1", 1.0], ["2002Q3", 2.0]],
            self.wb1.sheets[0]["A1"].options(expand="table").value,
        )

    def test_NA(self):
        df = pd.DataFrame([pd.NA, np.nan], index=[0.0, 1.0], columns=["one"])
        self.wb1.sheets[0].range("A1").value = df
        assert_frame_equal(
            df,
            self.wb1.sheets[0].range("A1").options(pd.DataFrame, expand="table").value,
        )


if __name__ == "__main__":
    unittest.main()



try:
    import pandas as pd
except ImportError:
    pd = None
try:
    import numpy as np
except ImportError:
    np = None

from .framework import (
    Accessor,
    ConversionContext,
    Converter,
    Options,
    Pipeline,
    accessors,
)
from .standard import (
    AdjustDimensionsStage,
    CleanDataForWriteStage,
    CleanDataFromReadStage,
    DictConverter,
    Ensure2DStage,
    ExpandRangeStage,
    RangeAccessor,
    RawValueAccessor,
    ReadValueFromRangeStage,
    TransposeStage,
    ValueAccessor,
    WriteValueToRangeStage,
)

if np:
    from .numpy_conv import NumpyArrayConverter
if pd:
    from .pandas_conv import PandasDataFrameConverter, PandasSeriesConverter

from .. import LicenseError

try:
    from ..pro.reports.markdown import Markdown, MarkdownConverter

    MarkdownConverter.register(Markdown)
except (ImportError, LicenseError, AttributeError):
    pass


__all__ = (
    "Accessor",
    "ConversionContext",
    "Converter",
    "Options",
    "Pipeline",
    "accessors",
    "AdjustDimensionsStage",
    "CleanDataForWriteStage",
    "CleanDataFromReadStage",
    "DictConverter",
    "Ensure2DStage",
    "ExpandRangeStage",
    "RangeAccessor",
    "RawValueAccessor",
    "ReadValueFromRangeStage",
    "TransposeStage",
    "ValueAccessor",
    "WriteValueToRangeStage",
    "NumpyArrayConverter",
    "PandasDataFrameConverter",
    "PandasSeriesConverter",
)


def read(rng, value, options, engine_name=None):
    convert = options.get("convert", None)
    pipeline = accessors.get(convert, convert).reader(options)
    ctx = ConversionContext(rng=rng, value=value, engine_name=engine_name)
    pipeline(ctx)
    return ctx.value


def write(value, rng, options, engine_name=None):
    # Don't allow to write lists and tuples as jagged arrays as appscript and pywin32
    # don't handle that properly. This should really be handled in Ensure2DStage, but
    # we'd have to set the original format in the conversion ctx meta as the check
    # should only run for list and tuples.
    if (
        isinstance(value, (list, tuple))
        and len(value) > 0
        and isinstance(value[0], (list, tuple))
    ):
        first_row = value[0]
        for row in value:
            if len(first_row) != len(row):
                raise Exception(
                    "All elements of a 2d list or tuple must be of the same length"
                )
    convert = options.get("convert", None)
    pipeline = (
        accessors.get(convert, convert).router(value, rng, options).writer(options)
    )
    ctx = ConversionContext(rng=rng, value=value, engine_name=engine_name)
    pipeline(ctx)
    return ctx.value



import os

import xlwings as xw

from .common import TestBase, this_dir

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None


class TestActive(TestBase):
    def test_apps_active(self):
        self.app2.activate()
        self.assertEqual(xw.apps.active, self.app2)

    def test_books_active(self):
        wb = xw.Book()
        self.assertEqual(xw.books.active, wb)

    def test_sheets_active(self):
        self.wb2.sheets[0].name = "active sheet test"
        self.assertEqual(self.wb2.sheets.active.name, "active sheet test")

    def test_range(self):
        xw.sheets.active.range("B2:C3").value = 123.0
        self.assertEqual(xw.Range("B2:C3").value, [[123.0, 123.0], [123.0, 123.0]])

    def test_book_fullname_closed(self):
        wb = xw.Book(os.path.join(this_dir, "test book.xlsx"))
        self.assertEqual(wb, xw.apps.active.books["test book.xlsx"])

    def test_book_fullname_open(self):
        wb1 = self.app1.books.open(os.path.join(this_dir, "test book.xlsx"))
        wb2 = xw.Book(os.path.join(this_dir, "test book.xlsx"))
        self.assertEqual(wb1, wb2)

    def test_book_name_closed(self):
        os.chdir(this_dir)
        wb = xw.Book("test book.xlsx")
        self.assertEqual(wb, xw.apps.active.books["test book.xlsx"])

    def test_book_name_open(self):
        wb1 = self.app1.books.open(os.path.join(this_dir, "test book.xlsx"))
        wb2 = xw.Book("test book.xlsx")
        self.assertEqual(wb1, wb2)

    def test_book_open_bad_name(self):
        with self.assertRaises(FileNotFoundError):
            xw.Book("bad name.xlsx")

    def test_book(self):
        wb = xw.Book()
        self.assertEqual(wb, xw.apps.active.books.active)

    def test_book_name_unsaved(self):
        wb = xw.Book()
        self.assertEqual(wb, xw.Book(wb.name))

    def test_books(self):
        wb = xw.Book()
        self.assertEqual(xw.books[-1], wb)


class TestView(TestBase):
    def test_list_new_book(self):
        n_books = xw.books.count
        xw.view([1, 2, 3])
        self.assertEqual(xw.books.count, n_books + 1)

    def test_list_sheet(self):
        n_books = xw.books.count
        xw.view([1, 2, 3], sheet=xw.books[0].sheets[0])
        self.assertEqual(xw.books.count, n_books)
        self.assertEqual(xw.books[0].sheets[0].range("A1:C1").value, [1.0, 2.0, 3.0])



import unittest

from .common import TestBase


class TestNames(TestBase):
    def test_get_names_index(self):
        self.wb1.sheets[0].range("B2:D10").name = "test1"
        self.wb1.sheets[0].range("A1").name = "test2"
        self.assertEqual(self.wb1.names(1).name, "test1")
        self.assertEqual(self.wb1.names[1].name, "test2")

    def test_names_contain(self):
        self.wb1.sheets[0].range("B2:D10").name = "test1"
        self.assertTrue("test1" in self.wb1.names)

    def test_len(self):
        self.wb1.sheets[0].range("B2:D10").name = "test1"
        self.wb1.sheets[0].range("A1").name = "test2"
        self.assertEqual(len(self.wb1.names), 2)

    def test_count(self):
        self.assertEqual(len(self.wb1.names), self.wb1.names.count)

    def test_names_iter(self):
        self.wb1.sheets[0].range("B2:D10").name = "test1"
        self.wb1.sheets[0].range("A1").name = "test2"
        for ix, n in enumerate(self.wb1.names):
            if ix == 0:
                self.assertEqual(n.name, "test1")
            if ix == 1:
                self.assertEqual(n.name, "test2")

    def test_get_inexisting_name(self):
        self.assertIsNone(self.wb1.sheets[0].range("A1").name)

    def test_get_set_named_range(self):
        self.wb1.sheets[0].range("A100").name = "test1"
        self.assertEqual(self.wb1.sheets[0].range("A100").name.name, "test1")

        self.wb1.sheets[0].range("A200:B204").name = "test2"
        self.assertEqual(self.wb1.sheets[0].range("A200:B204").name.name, "test2")

    def test_delete_named_item1(self):
        self.wb1.sheets[0].range("B10:C11").name = "to_be_deleted"
        self.assertEqual(
            self.wb1.sheets[0].range("to_be_deleted").name.name, "to_be_deleted"
        )

        del self.wb1.names["to_be_deleted"]
        self.assertIsNone(self.wb1.sheets[0].range("B10:C11").name)

    def test_delete_named_item2(self):
        self.wb1.sheets[0].range("B10:C11").name = "to_be_deleted"
        self.assertEqual(
            self.wb1.sheets[0].range("to_be_deleted").name.name, "to_be_deleted"
        )

        self.wb1.names["to_be_deleted"].delete()
        self.assertIsNone(self.wb1.sheets[0].range("B10:C11").name)

    def test_delete_named_item3(self):
        self.wb1.sheets[0].range("B10:C11").name = "to_be_deleted"
        self.assertEqual(
            self.wb1.sheets[0].range("to_be_deleted").name.name, "to_be_deleted"
        )

        self.wb1.sheets[0].range("to_be_deleted").name.delete()
        self.assertIsNone(self.wb1.sheets[0].range("B10:C11").name)

    def test_names_collection(self):
        self.wb1.sheets[0].range("A1").name = "name1"
        self.wb1.sheets[0].range("A2").name = "name2"
        self.assertTrue("name1" in self.wb1.names and "name2" in self.wb1.names)

        self.wb1.sheets[0].range("A3").name = "name3"
        self.assertTrue(
            "name1" in self.wb1.names
            and "name2" in self.wb1.names
            and "name3" in self.wb1.names
        )

    def test_sheet_scope(self):
        self.wb2.sheets[0].range("B2:C3").name = "Sheet1!sheet_scope1"
        self.wb2.sheets[0].range("sheet_scope1").value = [[1.0, 2.0], [3.0, 4.0]]
        self.assertEqual(
            self.wb2.sheets[0].range("B2:C3").value, [[1.0, 2.0], [3.0, 4.0]]
        )
        with self.assertRaises(Exception):
            values = self.wb2.sheets[1].range("sheet_scope1").value

    def test_workbook_scope(self):
        self.wb1.sheets[0].range("A1").name = "test1"
        self.wb1.sheets[0].range("test1").value = 123.0
        self.assertEqual(self.wb1.names["test1"].refers_to_range.value, 123.0)

    def test_contains_name(self):
        self.wb1.sheets[0].range("A1").name = "test1"
        self.assertTrue(self.wb1.names.contains("test1"))
        self.assertFalse(self.wb1.names.contains("test2"))

    def test_wb_names_add(self):
        self.wb1.names.add("test1", "=Sheet1!$A$1:$B$3")
        self.assertEqual(self.wb1.sheets[0].range("A1:B3").name.name, "test1")

    def test_sht_names_add(self):
        self.wb1.sheets[0].names.add("test1", "=Sheet1!$A$1:$B$3")
        self.assertEqual(self.wb1.sheets[0].range("A1:B3").name.name, "Sheet1!test1")

    def test_refers_to_range(self):
        self.wb1.sheets[0].range("B2:D10").name = "test1"
        self.assertEqual(
            self.wb1.sheets[0].range("B2:D10"),
            self.wb1.sheets[0].range("B2:D10").name.refers_to_range,
        )

    def test_refers_to_range_sheet_with_spaces(self):
        # This will cause quotes around sheet reference which caused a bug on mac
        self.wb1.sheets[0].name = "She et1"
        self.wb1.sheets[0].range("B2:D10").name = "test1"
        self.assertEqual(
            self.wb1.sheets["She et1"].range("B2:D10"),
            self.wb1.names["test1"].refers_to_range,
        )


if __name__ == "__main__":
    unittest.main()



from pathlib import Path

import pytest

import xlwings as xw

this_dir = Path(__file__).resolve().parent


@pytest.fixture(scope="module")
def app():
    with xw.App(visible=False) as app:
        app.books.open(this_dir / "test book.xlsx")
        yield app


# Range
def test_range_font_bold(app):
    sheet = app.books[0].sheets[0]
    sheet["A1"].value = "text"
    assert sheet["A1"].font.bold is False
    sheet["A1"].font.bold = True
    assert sheet["A1"].font.bold is True
    sheet["A1"].font.bold = False
    assert sheet["A1"].font.bold is False


def test_range_font_italic(app):
    sheet = app.books[0].sheets[0]
    sheet["A1"].value = "text"
    assert sheet["A1"].font.italic is False
    sheet["A1"].font.italic = True
    assert sheet["A1"].font.italic is True
    sheet["A1"].font.italic = False
    assert sheet["A1"].font.italic is False


def test_range_font_size(app):
    sheet = app.books[0].sheets[0]
    sheet["A1"].value = "text"
    assert sheet["A1"].font.size != 0
    sheet["A1"].font.size = 33
    assert sheet["A1"].font.size == 33


def test_range_font_color(app):
    sheet = app.books[0].sheets[0]
    sheet["A1"].value = "text"
    assert sheet["A1"].font.color == (0, 0, 0)
    sheet["A1"].font.color = (255, 0, 0)
    assert sheet["A1"].font.color == (255, 0, 0)


def test_range_font_name(app):
    sheet = app.books[0].sheets[0]
    sheet["A1"].value = "text"
    sheet["A1"].font.name = "Calibri"
    assert sheet["A1"].font.name == "Calibri"
    sheet["A1"].font.name = "Arial"
    assert sheet["A1"].font.name == "Arial"


# Shape
def test_shape_font_bold(app):
    shape = xw.Book("test book.xlsx").sheets["shape"].shapes[0]
    shape.text = "text"
    assert shape.font.bold is False
    shape.font.bold = True
    assert shape.font.bold is True
    shape.font.bold = False
    assert shape.font.bold is False


def test_shape_font_italic(app):
    shape = xw.Book("test book.xlsx").sheets["shape"].shapes[0]
    shape.text = "text"
    assert shape.font.italic is False
    shape.font.italic = True
    assert shape.font.italic is True
    shape.font.italic = False
    assert shape.font.italic is False


def test_shape_font_size(app):
    shape = xw.Book("test book.xlsx").sheets["shape"].shapes[0]
    shape.text = "text"
    assert shape.font.size != 33
    shape.font.size = 33
    assert shape.font.size == 33


def test_shape_font_color(app):
    shape = xw.Book("test book.xlsx").sheets["shape"].shapes[0]
    shape.text = "text"
    assert shape.font.color == (255, 255, 255)
    shape.font.color = (255, 0, 0)
    assert shape.font.color == (255, 0, 0)


def test_shape_font_name(app):
    shape = xw.Book("test book.xlsx").sheets["shape"].shapes[0]
    shape.text = "text"
    shape.font.name = "Calibri"
    assert shape.font.name == "Calibri"
    shape.font.name = "Arial"
    assert shape.font.name == "Arial"



import datetime as dt
import os
import re
import subprocess
import sys
import tempfile
import traceback
import uuid
from functools import lru_cache, total_ordering
from pathlib import Path

try:
    import numpy as np
except ImportError:
    np = None

try:
    import matplotlib as mpl
    import matplotlib.figure  # noqa: F401
    import matplotlib.pyplot as plt
except ImportError:
    mpl = None

try:
    import plotly.graph_objects as plotly_go
except ImportError:
    plotly_go = None

import xlwings

missing = object()


def int_to_rgb(number):
    """Given an integer, return the rgb"""
    number = int(number)
    r = number % 256
    g = (number // 256) % 256
    b = (number // (256 * 256)) % 256
    return r, g, b


def rgb_to_int(rgb):
    """Given an rgb, return an int"""
    return rgb[0] + (rgb[1] * 256) + (rgb[2] * 256 * 256)


def hex_to_rgb(color):
    color = color[1:] if color.startswith("#") else color
    return tuple(int(color[i : i + 2], 16) for i in (0, 2, 4))


def rgb_to_hex(r, g, b):
    return f"#{r:02x}{g:02x}{b:02x}"


def get_duplicates(seq):
    seen = set()
    duplicates = set(x for x in seq if x in seen or seen.add(x))
    return duplicates


def np_datetime_to_datetime(np_datetime):
    ts = (np_datetime - np.datetime64("1970-01-01T00:00:00Z")) / np.timedelta64(1, "s")
    dt_datetime = dt.datetime.utcfromtimestamp(ts)
    return dt_datetime


def xlserial_to_datetime(serial):
    """
    Converts a date in Excel's serial format (e.g., 44197.0) to a Python datetime object
    """
    # https://learn.microsoft.com/en-us/office/dev/scripts/resources/samples/excel-samples#dates
    ts = round((serial - 25569) * 86400, 3)
    return dt.datetime.utcfromtimestamp(ts)  # tz-naive, which is what we want


def datetime_to_xlserial(obj):
    """
    Converts a Python date or datetime object to Excel's date serial (e.g, 44197.0)
    """
    if isinstance(obj, dt.datetime):
        obj = obj.replace(tzinfo=dt.timezone.utc)
    elif isinstance(obj, dt.date):
        obj = dt.datetime(
            obj.year,
            obj.month,
            obj.day,
            tzinfo=dt.timezone.utc,
        )
    return obj.timestamp() / 86400 + 25569


ALPHABET = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"


def col_name(i):
    i -= 1
    if i < 0:
        raise IndexError(i)
    elif i < 26:
        return ALPHABET[i]
    elif i < 702:
        i -= 26
        return ALPHABET[i // 26] + ALPHABET[i % 26]
    elif i < 16384:
        i -= 702
        return ALPHABET[i // 676] + ALPHABET[i // 26 % 26] + ALPHABET[i % 26]
    else:
        raise IndexError(i)


def address_to_index_tuple(address):
    """
    Based on a function from XlsxWriter, which is distributed under the following
    BSD 2-Clause License:

    Copyright (c) 2013-2021, John McNamara <jmcnamara@cpan.org>
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice, this
       list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
    SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    """
    re_range_parts = re.compile(r"(\$?)([A-Z]{1,3})(\$?)(\d+)")
    match = re_range_parts.match(address)
    if match:
        col_str = match.group(2)
        row_str = match.group(4)

        # Convert base26 column string to number
        expn = 0
        col = 0
        for char in reversed(col_str):
            col += (ord(char) - ord("A") + 1) * (26**expn)
            expn += 1

        return int(row_str), col


def a1_to_tuples(address):
    if ":" in address:
        address1, address2 = address.split(":")
        tuple1 = address_to_index_tuple(address1.upper())
        tuple2 = address_to_index_tuple(address2.upper())
    else:
        tuple1 = address_to_index_tuple(address.upper())
        tuple2 = None
    return tuple1, tuple2


class VBAWriter:
    MAX_VBA_LINE_LENGTH = 1024
    VBA_LINE_SPLIT = " _\n"
    MAX_VBA_SPLITTED_LINE_LENGTH = MAX_VBA_LINE_LENGTH - len(VBA_LINE_SPLIT)

    class Block:
        def __init__(self, writer, start):
            self.writer = writer
            self.start = start

        def __enter__(self):
            self.writer.writeln(self.start)
            self.writer._indent += 1

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.writer._indent -= 1

    def __init__(self, f):
        self.f = f
        self._indent = 0
        self._freshline = True

    def block(self, template, **kwargs):
        return VBAWriter.Block(self, template.format(**kwargs))

    def start_block(self, template, **kwargs):
        self.writeln(template, **kwargs)
        self._indent += 1

    def end_block(self, template, **kwargs):
        self.writeln(template, **kwargs)
        self._indent -= 1

    def write(self, template, **kwargs):
        if kwargs:
            template = template.format(**kwargs)
        if self._freshline:
            template = ("    " * self._indent) + template
            self._freshline = False
        self.write_vba_line(template)
        if template[-1] == "\n":
            self._freshline = True

    def write_label(self, label):
        self._indent -= 1
        self.write(label + ":\n")
        self._indent += 1

    def writeln(self, template, **kwargs):
        self.write(template + "\n", **kwargs)

    def write_vba_line(self, vba_line):
        if len(vba_line) > VBAWriter.MAX_VBA_LINE_LENGTH:
            separator_index = VBAWriter.get_separator_index(vba_line)
            self.f.write(vba_line[:separator_index] + VBAWriter.VBA_LINE_SPLIT)
            self.write_vba_line(vba_line[separator_index:])
        else:
            self.f.write(vba_line)

    @classmethod
    def get_separator_index(cls, vba_line):
        for index in range(cls.MAX_VBA_SPLITTED_LINE_LENGTH, 0, -1):
            if " " == vba_line[index]:
                return index
        return (
            cls.MAX_VBA_SPLITTED_LINE_LENGTH
        )  # Best effort: split string at the maximum possible length


def try_parse_int(x):
    try:
        return int(x)
    except ValueError:
        return x


@total_ordering
class VersionNumber:
    def __init__(self, s):
        self.value = tuple(map(try_parse_int, s.split(".")))

    @property
    def major(self):
        return self.value[0]

    @property
    def minor(self):
        return self.value[1] if len(self.value) > 1 else None

    def __str__(self):
        return ".".join(map(str, self.value))

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, repr(str(self)))

    def __eq__(self, other):
        if isinstance(other, VersionNumber):
            return self.value == other.value
        elif isinstance(other, str):
            return self.value == VersionNumber(other).value
        elif isinstance(other, tuple):
            return self.value[: len(other)] == other
        elif isinstance(other, int):
            return self.major == other
        else:
            return False

    def __lt__(self, other):
        if isinstance(other, VersionNumber):
            return self.value < other.value
        elif isinstance(other, str):
            return self.value < VersionNumber(other).value
        elif isinstance(other, tuple):
            return self.value[: len(other)] < other
        elif isinstance(other, int):
            return self.major < other
        else:
            raise TypeError("Cannot compare other object with version number")


def process_image(image, format, export_options):
    """Returns filename and is_temp_file"""
    image = fspath(image)
    if isinstance(image, str):
        return image, False
    elif mpl and isinstance(image, mpl.figure.Figure):
        image_type = "mpl"
    elif plotly_go and isinstance(image, plotly_go.Figure):
        image_type = "plotly"
    else:
        raise TypeError("Don't know what to do with that image object")

    if export_options is None:
        export_options = {"bbox_inches": "tight", "dpi": 200}

    if format == "vector":
        if sys.platform.startswith("darwin"):
            format = "pdf"
        else:
            format = "svg"

    temp_dir = os.path.realpath(tempfile.gettempdir())
    filename = os.path.join(temp_dir, str(uuid.uuid4()) + "." + format)

    if image_type == "mpl":
        canvas = mpl.backends.backend_agg.FigureCanvas(image)
        canvas.draw()
        image.savefig(filename, **export_options)
        plt.close(image)
    elif image_type == "plotly":
        image.write_image(filename)
    return filename, True


def fspath(path):
    """Convert path-like object to string.

    On python <= 3.5 the input argument is always returned unchanged (no support for
    path-like objects available). TODO: can be removed as 3.5 no longer supported.

    """
    if hasattr(os, "PathLike") and isinstance(path, os.PathLike):
        return os.fspath(path)
    else:
        return path


def read_config_sheet(book):
    try:
        return book.sheets["xlwings.conf"]["A1:B1"].options(dict, expand="down").value
    except:  # noqa: E722
        # A missing sheet currently produces different errors on mac and win
        return {}


def read_user_config():
    """Returns keys in lowercase of xlwings.conf in the user's home directory"""
    config = {}
    if Path(xlwings.USER_CONFIG_FILE).is_file():
        with open(xlwings.USER_CONFIG_FILE, "r") as f:
            for line in f:
                values = re.findall(r'"[^"]*"', line)
                if values:
                    config[values[0].strip('"').lower()] = os.path.expandvars(
                        values[1].strip('"')
                    )
    return config


@lru_cache(None)
def get_cached_user_config(key):
    return read_user_config().get(key.lower())


def exception(logger, msg, *args):
    if logger.hasHandlers():
        logger.exception(msg, *args)
    else:
        print(msg % args)
        traceback.print_exc()


def chunk(sequence, chunksize):
    for i in range(0, len(sequence), chunksize):
        yield sequence[i : i + chunksize]


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
            It must be "yes" (the default), "no" or None (meaning
            an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".

    Licensed under the MIT License
    Copyright by Trent Mick
    https://code.activestate.com/recipes/577058/
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == "":
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n")


def prepare_sys_path(args_string):
    """Called from Excel to prepend the default paths and those from the PYTHONPATH
    setting to sys.path. While RunPython could use Book.caller(), the UDF server can't,
    as this runs before VBA can push the ActiveWorkbook over. UDFs also can't interact
    with the book object in general as Excel is busy during the function call and so
    won't allow you to read out the config sheet, for example. Before 0.24.9,
    these manipulations were handled in VBA, but couldn't handle SharePoint.
    """
    args = os.path.normcase(os.path.expandvars(args_string)).split(";")
    paths = []
    if args[0].lower() == "true":  # Add dir of Excel file to PYTHONPATH
        # Not sure if we really need normcase,
        # but on Windows it replaces "/" with "\", so let's revert that
        active_fullname = args[1].replace("\\", "/")
        this_fullname = args[2].replace("\\", "/")
        for fullname in [active_fullname, this_fullname]:
            if not fullname:
                continue
            elif "://" in fullname:
                fullname = Path(
                    fullname_url_to_local_path(
                        url=fullname,
                        sheet_onedrive_consumer_config=args[3],
                        sheet_onedrive_commercial_config=args[4],
                        sheet_sharepoint_config=args[5],
                    )
                )
            else:
                fullname = Path(fullname)
            paths += [str(fullname.parent), str(fullname.with_suffix(".zip"))]

    if args[6:]:
        paths += args[6:]

    if paths:
        sys.path[0:0] = list(set(paths))


@lru_cache(None)
def fullname_url_to_local_path(
    url,
    sheet_onedrive_consumer_config=None,
    sheet_onedrive_commercial_config=None,
    sheet_sharepoint_config=None,
):
    """
    When AutoSave is enabled in Excel with either OneDrive or SharePoint, VBA/COM's
    Workbook.FullName turns into a URL without any possibilities to get the local file
    path. While OneDrive and OneDrive for Business make it easy enough to derive the
    local path from the URL, SharePoint allows to define the "Site name" and "Site
    address" independently from each other with the former ending up in the local folder
    path and the latter in the FullName URL. Adding to the complexity: (1) When the site
    name contains spaces, they will be stripped out from the URL and (2) you can sync a
    subfolder directly (this, at least, works when you have a single folder at the
    SharePoint's Document root), which results in skipping a folder level locally when
    compared to the online/URL version. And (3) the OneDriveCommercial env var sometimes
    seems to actually point to the local SharePoint folder.

    Parameters
    ----------
    url : str
        URL as returned by VBA's FullName

    sheet_onedrive_consumer_config : str
        Optional Path to the local OneDrive (Personal) as defined in the Workbook's
        config sheet

    sheet_onedrive_commercial_config : str
        Optional Path to the local OneDrive for Business as defined in the Workbook's
        config sheet

    sheet_sharepoint_config : str
        Optional Path to the local SharePoint drive as defined in the Workbook's config
        sheet
    """
    # Directory config files can't be used
    # since the whole purpose of this exercise is to find out about a book's dir
    onedrive_consumer_config_name = (
        "ONEDRIVE_CONSUMER_WIN"
        if sys.platform.startswith("win")
        else "ONEDRIVE_CONSUMER_MAC"
    )
    onedrive_commercial_config_name = (
        "ONEDRIVE_COMMERCIAL_WIN"
        if sys.platform.startswith("win")
        else "ONEDRIVE_COMMERCIAL_MAC"
    )
    sharepoint_config_name = (
        "SHAREPOINT_WIN" if sys.platform.startswith("win") else "SHAREPOINT_MAC"
    )
    if sheet_onedrive_consumer_config is not None:
        sheet_onedrive_consumer_config = os.path.expandvars(
            sheet_onedrive_consumer_config
        )
    if sheet_onedrive_commercial_config is not None:
        sheet_onedrive_commercial_config = os.path.expandvars(
            sheet_onedrive_commercial_config
        )
    if sheet_sharepoint_config is not None:
        sheet_sharepoint_config = os.path.expandvars(sheet_sharepoint_config)
    onedrive_consumer_config = sheet_onedrive_consumer_config or read_user_config().get(
        onedrive_consumer_config_name.lower()
    )
    onedrive_commercial_config = (
        sheet_onedrive_commercial_config
        or read_user_config().get(onedrive_commercial_config_name.lower())
    )
    sharepoint_config = sheet_sharepoint_config or read_user_config().get(
        sharepoint_config_name.lower()
    )

    # OneDrive
    pattern = re.compile(r"https://d.docs.live.net/[^/]*/(.*)")
    match = pattern.match(url)
    if match:
        if sys.platform.startswith("darwin"):
            default_dir = Path.home() / "Library" / "CloudStorage" / "OneDrive-Personal"
        else:
            default_dir = Path.home() / "OneDrive"
        legacy_default_dir = Path.home() / "OneDrive"
        root = (
            onedrive_consumer_config
            or os.getenv("OneDriveConsumer")
            or os.getenv("OneDrive")
            or (str(default_dir) if default_dir.is_dir() else None)
            or (str(legacy_default_dir) if legacy_default_dir else None)
        )
        if not root:
            raise xlwings.XlwingsError(
                f"Couldn't find the local OneDrive folder. Please configure the "
                f"{onedrive_consumer_config_name} setting, see: xlwings.org/error."
            )
        local_path = Path(root) / match.group(1)
        if local_path.is_file():
            return str(local_path)
        else:
            raise xlwings.XlwingsError(
                "Couldn't find your local OneDrive file, see: xlwings.org/error"
            )

    # OneDrive for Business
    pattern = re.compile(r"https://[^-]*-my.sharepoint.[^/]*/[^/]*/[^/]*/[^/]*/(.*)")
    match = pattern.match(url)
    if match:
        root = (
            onedrive_commercial_config
            or os.getenv("OneDriveCommercial")
            or os.getenv("OneDrive")
        )
        if not root:
            raise xlwings.XlwingsError(
                f"Couldn't find the local OneDrive for Business folder. "
                f"Please configure the {onedrive_commercial_config_name} setting, "
                f"see: xlwings.org/error."
            )
        local_path = Path(root) / match.group(1)
        if local_path.is_file():
            return str(local_path)
        else:
            raise xlwings.XlwingsError(
                "Couldn't find your local OneDrive for Business file, "
                "see: xlwings.org/error"
            )
    # SharePoint Online & On-Premises (Windows registry)
    url_to_mount = get_url_to_mount()
    for url_namespace, mount_point in url_to_mount.items():
        if url.startswith(url_namespace):
            local_path = Path(mount_point) / url[len(url_namespace) :]
            if local_path.is_file():
                return str(local_path)
            else:
                return search_local_sharepoint_path(
                    url, mount_point, sharepoint_config, sharepoint_config_name
                )
    # SharePoint Online & On-Premises (default top level mapping)
    pattern = re.compile(r"https?://[^/]*/sites/([^/]*)/([^/]*)/(.*)")
    match = pattern.match(url)
    # We're trying to derive the SharePoint root path
    # from the OneDriveCommercial path, if it exists
    root = sharepoint_config or (
        os.getenv("OneDriveCommercial").replace("OneDrive - ", "")
        if os.getenv("OneDriveCommercial")
        else None
    )
    if not root:
        raise xlwings.XlwingsError(
            f"Couldn't find the local SharePoint folder. Please configure the "
            f"{sharepoint_config_name} setting, see: xlwings.org/error."
        )
    if match:
        local_path = Path(root) / f"{match.group(1)} - Documents" / match.group(3)
        if local_path.is_file():
            return str(local_path)
    # SharePoint Online & On-Premises (non-default mapping)
    return search_local_sharepoint_path(
        url, root, sharepoint_config, sharepoint_config_name
    )


def to_pdf(
    obj,
    path=None,
    include=None,
    exclude=None,
    layout=None,
    exclude_start_string=None,
    show=None,
    quality=None,
):
    report_path = fspath(path)
    layout_path = fspath(layout)
    if isinstance(obj, (xlwings.Book, xlwings.Sheet)):
        if report_path is None:
            filename, extension = os.path.splitext(obj.fullname)
            directory, _ = os.path.split(obj.fullname)
            if directory:
                report_path = os.path.join(directory, filename + ".pdf")
            else:
                report_path = filename + ".pdf"
        if (include is not None) and (exclude is not None):
            raise ValueError("You can only use either 'include' or 'exclude'")
        # Hide sheets to exclude them from printing
        if isinstance(include, (str, int)):
            include = [include]
        if isinstance(exclude, (str, int)):
            exclude = [exclude]
        exclude_by_name = [
            sheet.index
            for sheet in obj.sheets
            if sheet.name.startswith(exclude_start_string)
        ]
        visibility = {}
        if include or exclude or exclude_by_name:
            for sheet in obj.sheets:
                visibility[sheet] = sheet.visible
        try:
            if include:
                for sheet in obj.sheets:
                    if (sheet.name in include) or (sheet.index in include):
                        sheet.visible = True
                    else:
                        sheet.visible = False
            if exclude or exclude_by_name:
                exclude = [] if exclude is None else exclude
                for sheet in obj.sheets:
                    if (
                        (sheet.name in exclude)
                        or (sheet.index in exclude)
                        or (sheet.index in exclude_by_name)
                    ):
                        sheet.visible = False
            obj.impl.to_pdf(os.path.realpath(report_path), quality=quality)
        except Exception:
            raise
        finally:
            # Reset visibility
            if include or exclude or exclude_by_name:
                for sheet, tf in visibility.items():
                    sheet.visible = tf
    else:
        if report_path is None:
            if isinstance(obj, xlwings.Chart):
                directory, _ = os.path.split(obj.parent.book.fullname)
                filename = obj.name
            elif isinstance(obj, xlwings.Range):
                directory, _ = os.path.split(obj.sheet.book.fullname)
                filename = (
                    str(obj)
                    .replace("<", "")
                    .replace(">", "")
                    .replace(":", "_")
                    .replace(" ", "")
                )
            else:
                raise ValueError(f"Object of type {type(obj)} are not supported.")
            if directory:
                report_path = os.path.join(directory, filename + ".pdf")
            else:
                report_path = filename + ".pdf"
        obj.impl.to_pdf(os.path.realpath(report_path), quality=quality)

    if layout:
        from .pro.reports.pdf import print_on_layout

        print_on_layout(report_path=report_path, layout_path=layout_path)

    if show:
        if sys.platform.startswith("win"):
            os.startfile(report_path)
        else:
            subprocess.run(["open", report_path])
    return report_path


def get_url_to_mount():
    """Windows stores the mount points in the registry. This helps but still isn't
    foolproof.
    """
    if sys.platform.startswith("win"):
        import winreg
        from winreg import HKEY_CURRENT_USER, KEY_READ

        root = r"SOFTWARE\SyncEngines\Providers\OneDrive"
        url_to_mount = {}
        try:
            with winreg.OpenKey(HKEY_CURRENT_USER, root, 0, KEY_READ) as root_key:
                for i in range(0, winreg.QueryInfoKey(root_key)[0]):
                    subfolder = winreg.EnumKey(root_key, i)
                    with winreg.OpenKey(
                        HKEY_CURRENT_USER, f"{root}\\{subfolder}", 0, KEY_READ
                    ) as key:
                        try:
                            mount_point, _ = winreg.QueryValueEx(key, "MountPoint")
                            url_namespace, _ = winreg.QueryValueEx(key, "URLNamespace")
                            url_to_mount[url_namespace] = mount_point
                        except FileNotFoundError:
                            pass
        except FileNotFoundError:
            pass
        return url_to_mount
    else:
        return {}


def search_local_sharepoint_path(url, root, sharepoint_config, sharepoint_config_name):
    book_name = url.split("/")[-1]
    local_book_paths = []
    for path in Path(root).rglob("[!~$]*.xls*"):
        if path.name.lower() == book_name.lower():
            local_book_paths.append(path)
    if len(local_book_paths) == 1:
        return str(local_book_paths[0])
    elif len(local_book_paths) == 0:
        raise xlwings.XlwingsError(
            "Couldn't find your SharePoint file locally, see: xlwings.org/error"
        )
    else:
        raise xlwings.XlwingsError(
            f"Your SharePoint configuration either requires your workbook name to be "
            f"unique across all synced SharePoint folders or you need to "
            f"{'edit' if sharepoint_config else 'add'} the {sharepoint_config_name} "
            f"setting including one or more folder levels, see: xlwings.org/error."
        )


def excel_update_picture(picture_impl, filename):
    name = picture_impl.name
    left, top = picture_impl.left, picture_impl.top
    width, height = picture_impl.width, picture_impl.height

    picture_impl.delete()

    picture_impl = picture_impl.parent.pictures.add(
        filename,
        link_to_file=False,
        save_with_document=True,
        left=left,
        top=top,
        width=width,
        height=height,
        anchor=None,
    )

    picture_impl.name = name
    return picture_impl



"""See https://github.com/xlwings/xlwings/issues/1789
Can be removed again if there's a solution for
https://github.com/mhammond/pywin32/issues/1870

This file's content is taken from pywin32 v301,
distributed under the following license:

Unless stated in the specfic source file, this work is
Copyright (c) 1996-2008, Greg Stein and Mark Hammond.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in
the documentation and/or other materials provided with the distribution.

Neither names of Greg Stein, Mark Hammond nor the name of contributors may be used
to endorse or promote products derived from this software without
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import pythoncom


class CoClassBaseClass:
    def __init__(self, oobj=None):
        if oobj is None:
            oobj = pythoncom.new(self.CLSID)
        self.__dict__["_dispobj_"] = self.default_interface(oobj)

    def __repr__(self):
        return "<win32com.gen_py.%s.%s>" % (__doc__, self.__class__.__name__)

    def __getattr__(self, attr):
        d = self.__dict__["_dispobj_"]
        if d is not None:
            return getattr(d, attr)
        raise AttributeError(attr)

    def __setattr__(self, attr, value):
        if attr in self.__dict__:
            self.__dict__[attr] = value
            return
        try:
            d = self.__dict__["_dispobj_"]
            if d is not None:
                d.__setattr__(attr, value)
                return
        except AttributeError:
            pass
        self.__dict__[attr] = value

    # Special methods don't use __getattr__ etc, so explicitly delegate here.
    # Some wrapped objects might not have them, but that's OK - the attribute
    # error can just bubble up.
    def __call__(self, *args, **kwargs):
        return self.__dict__["_dispobj_"].__call__(*args, **kwargs)

    def __str__(self, *args):
        return self.__dict__["_dispobj_"].__str__(*args)

    def __int__(self, *args):
        return self.__dict__["_dispobj_"].__int__(*args)

    def __iter__(self):
        return self.__dict__["_dispobj_"].__iter__()

    def __len__(self):
        return self.__dict__["_dispobj_"].__len__()

    def __nonzero__(self):
        return self.__dict__["_dispobj_"].__nonzero__()



"""
Required Notice: Copyright (C) Zoomer Analytics GmbH.

xlwings PRO is dual-licensed under one of the following licenses:

* PolyForm Noncommercial License 1.0.0 (for noncommercial use):
  https://polyformproject.org/licenses/noncommercial/1.0.0
* xlwings PRO License (for commercial use):
  https://github.com/xlwings/xlwings/blob/main/LICENSE_PRO.txt

Commercial licenses can be purchased at https://www.xlwings.org
"""

import json
import os
import sys
from functools import lru_cache
from pathlib import Path

from ..main import Book
from ..utils import get_cached_user_config, read_config_sheet
from .module_permissions import verify_execute_permission
from .utils import LicenseHandler, get_embedded_code_temp_dir

LicenseHandler.validate_license("pro")

TEMPDIR = get_embedded_code_temp_dir()


@lru_cache()
def dump_embedded_code(book, target_dir):
    code_map = read_config_sheet(book).get("RELEASE_EMBED_CODE_MAP", "{}")
    sheetname_to_path = json.loads(code_map)
    for sheet in book.sheets:
        if sheet.name.endswith(".py"):
            last_cell = sheet.used_range.last_cell
            sheet_content = (
                sheet.range((1, 1), (last_cell.row, 1)).options(ndim=1).value
            )
            if sheetname_to_path:
                (Path(target_dir) / sheetname_to_path[sheet.name]).parent.mkdir(
                    exist_ok=True
                )
            with open(
                os.path.join(
                    target_dir,
                    sheetname_to_path[sheet.name] if sheetname_to_path else sheet.name,
                ),
                "w",
                encoding="utf-8",
                newline="\n",
            ) as f:
                for row in sheet_content:
                    if row is None:
                        f.write("\n")
                    else:
                        f.write(row + "\n")
    sys.path[0:0] = [target_dir]


def runpython_embedded_code(command):
    dump_embedded_code(Book.caller(), TEMPDIR)
    if (
        get_cached_user_config("permission_check_enabled")
        and get_cached_user_config("permission_check_enabled").lower() == "true"
    ):
        verify_execute_permission(command=command)
    exec(command)



import datetime as dt
from pathlib import Path

import custom_functions
import jinja2
import markupsafe
from dateutil import tz
from fastapi import Body, FastAPI, Request, status
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import HTMLResponse, PlainTextResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

import xlwings as xw

# from tests import udf_tests_officejs as custom_functions

app = FastAPI()

this_dir = Path(__file__).resolve().parent


@app.post("/hello")
def hello(data: dict = Body):
    book = xw.Book(json=data)
    sheet = book.sheets[0]
    cell = sheet["A1"]
    if cell.value == "Hello xlwings!":
        cell.value = "Bye xlwings!"
    else:
        cell.value = "Hello xlwings!"
    return book.json()


@app.post("/show-alert")
def show_alert(data: dict = Body):
    book = xw.Book(json=data)
    book.app.alert(
        prompt="Some text",
        title="Some Title",
        buttons="ok_cancel",
        mode="info",
        callback="myCallback",
    )
    return book.json()


@app.post("/integration-test-read")
def integration_test_read(data: dict = Body):
    book = xw.Book(json=data)
    assert book.name == "engines.xlsm", "engines.xlsm must be the active file"
    if data["client"] == "Office.js":
        expected_data = expected_body["Office.js"]
    elif data["client"] == "VBA":
        expected_data = expected_body["VBA"]
    elif data["client"] == "Google Apps Script":
        expected_data = expected_body["Google Apps Script"]
    elif data["client"] == "Microsoft Office Scripts":
        expected_data = expected_body["Office Scripts"]
    assert data == expected_data, "Body differs (Make sure to select cell 'Sheet 1'!A1)"
    book.app.alert("OK", title="Integration Test Read")
    return book.json()


@app.post("/integration-test-write")
def integration_test_write(data: dict = Body):
    book = xw.Book(json=data)
    assert (
        book.name == "integration_write.xlsm"
    ), "integration_write.xlsm must be the active file"
    sheet1 = book.sheets["Sheet 1"]

    # Values
    sheet1["E3"].value = [
        [None, "string"],
        [-1, 1],
        [-1.1, 1.1],
        [True, False],
        [
            dt.date(2021, 7, 1),
            dt.datetime(2021, 12, 31, 23, 35, 12, tzinfo=tz.gettz("Europe/Paris")),
        ],
    ]

    # Add sheets and write to them
    # NOTE: in Excel Online, adding/renaming sheets makes an alert impossible to quit
    # via provided buttons ("osfControl for the given ID doesn't exist.")
    sheet2 = book.sheets.add("New Named Sheet")
    sheet2["A1"].value = "Named Sheet"
    sheet3 = book.sheets.add()
    sheet3["A1"].value = "Unnamed Sheet"

    # Tables
    if data["client"] != "Google Apps Script":
        sheet_tables = book.sheets["Tables"]

        sheet_tables["A1"].value = [["one", "two"], [1, 2], [3, 4]]
        sheet_tables.tables.add(sheet3["A1:B3"])

        sheet_tables["A5"].value = [[1, 2], [3, 4]]
        sheet_tables.tables.add(sheet_tables["A5:B6"], has_headers=False)

        sheet_tables["A9"].value = [["one", "two"], [1, 2], [3, 4]]
        mytable1 = sheet_tables.tables.add(sheet_tables["A9:B11"], name="MyTable1")
        mytable1.show_autofilter = False

        sheet_tables["A13"].value = [[1, 2], [3, 4]]
        mytable2 = sheet_tables.tables.add(
            sheet_tables["A13:B14"], name="MyTable2", has_headers=False
        )
        mytable2.show_headers = False
        mytable2.show_totals = True
        mytable2.show_filters = False
        mytable2.resize(sheet_tables["A14:C17"])

    # Set sheet name
    book.sheets["Sheet2"].name = "Changed"
    book.sheets["Changed"]["A1"].value = "Changed"

    # Autofit
    autofit_sheet = book.sheets["Autofit"]
    autofit_sheet["A1"].value = [[1, 2], [3, 4]]
    autofit_sheet["A1:B2"].autofit()
    autofit_sheet["D4:E5"].rows.autofit()
    autofit_sheet["G7:H82"].columns.autofit()

    # Range color
    sheet1["E12:F12"].color = "#3DBAC1"

    # Add Hyperlink
    sheet1["E14"].add_hyperlink("https://www.xlwings.org", "xlwings", "xw homepage")

    # Number format
    sheet1["E9:F10"].value = [[1, 2], [3, 4]]
    sheet1["E9:F10"].number_format = "0%"

    # Clear contents
    sheet1["E16:F17"].clear_contents()

    # Activate sheet
    book.sheets[1].activate()

    # Pictures
    sheet1.pictures.add(
        this_dir.parent / "tests" / "sample_picture.png",
        name="MyPic",
        anchor=sheet1["C28"],
    )
    sheet1.pictures.add(this_dir / "icons" / "icon-80.png", name="MyPic", update=True)
    book.sheets[1].pictures.add(this_dir.parent / "tests" / "sample_picture.png")

    # Add named ranges
    book.names.add("test1", "='Sheet 1'!$A$1:$B$3")
    book.names.add("test2", "=Changed!$A$1")
    sheet1["A1"].name = "test3"
    if data["client"] != "Google Apps Script":
        sheet1.names.add("test4", "='Sheet 1'!$A$1:$B$3")

    # Delete named ranges
    book.names["DeleteMe"].delete()
    book.names["Sheet4!DeleteMe"].delete()
    book.names["'Sheet 3'!DeleteMe"].delete()

    return book.json()


@app.get("/xlwings/alert", response_class=HTMLResponse)
async def alert(
    request: Request, prompt: str, title: str, buttons: str, mode: str, callback: str
):
    """This endpoint is required by myapp.alert()"""
    return templates.TemplateResponse(
        "xlwings-alert.html",
        {
            "request": request,
            "prompt": markupsafe.Markup(prompt.replace("\n", "<br>")),
            "title": title,
            "buttons": buttons,
            "mode": mode,
            "callback": callback,
        },
    )


@app.get("/xlwings/custom-functions-meta")
async def custom_functions_meta():
    return xw.pro.custom_functions_meta(custom_functions)


@app.get("/xlwings/custom-functions-code")
async def custom_functions_code():
    return PlainTextResponse(xw.pro.custom_functions_code(custom_functions))


@app.post("/xlwings/custom-functions-call")
async def custom_functions_call(request: Request, data: dict = Body):
    print(request.headers["Authorization"])
    rv = await xw.pro.custom_functions_call(data, custom_functions)
    return {"result": rv}


app.mount("/icons", StaticFiles(directory=this_dir / "icons"), name="icons")
app.mount("/", StaticFiles(directory=this_dir / "build"), name="home")
StaticFiles.is_not_modified = lambda *args, **kwargs: False  # Never cache static files

# Add the xlwings alert template as source
loader = jinja2.ChoiceLoader(
    [
        jinja2.FileSystemLoader(this_dir / "build"),
        jinja2.PackageLoader("xlwings", "html"),
    ]
)
templates = Jinja2Templates(directory=this_dir / "build", loader=loader)


# Office Scripts requires CORS and the following would be enough:
# allow_origin_regex=r"https://.*.officescripts.microsoftusercontent.com"
# but Office.js from Excel on the web also requires CORS and will have other origins
app.add_middleware(
    CORSMiddleware,
    allow_origins="*",
    allow_methods=["POST"],
    allow_headers=["*"],
)


@app.exception_handler(Exception)
async def exception_handler(request, exception):
    # Handling all Exceptions is OK since it's only a dev server, but you probably
    # don't want to show the details of every Exception to the user in production
    return PlainTextResponse(
        repr(exception), status_code=status.HTTP_500_INTERNAL_SERVER_ERROR
    )


expected_body = {}
expected_body["Office.js"] = {
    "client": "Office.js",
    "version": "dev",
    "book": {"name": "engines.xlsm", "active_sheet_index": 0, "selection": "A1"},
    "names": [
        {
            "name": "one",
            "sheet_index": 0,
            "address": "A1",
            "scope_sheet_name": None,
            "scope_sheet_index": None,
            "book_scope": True,
        },
        {
            "name": "two",
            "sheet_index": 1,
            "address": "A1:A2",
            "scope_sheet_name": None,
            "scope_sheet_index": None,
            "book_scope": True,
        },
        {
            "name": "two",
            "sheet_index": 0,
            "address": "C7:D8",
            "scope_sheet_name": "Sheet 1",
            "scope_sheet_index": 0,
            "book_scope": False,
        },
        {
            "name": "two",
            "sheet_index": 2,
            "address": "B3",
            "scope_sheet_name": "Sheet2",
            "scope_sheet_index": 1,
            "book_scope": False,
        },
    ],
    "sheets": [
        {
            "name": "Sheet 1",
            "values": [
                ["a", "b", "c", ""],
                [1.1, 2.2, 3.3, "2021-01-01T00:00:00.000Z"],
                [4.4, 5.5, 6.6, ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["Column1", "Column2", "", ""],
                [1.1, 2.2, "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                [1.1, 2.2, 3.3, ""],
                [4.4, 5.5, 6.6, ""],
                ["Total", "", 9.9, ""],
            ],
            # Width/height differ for Desktop and Web
            "pictures": [
                {"name": "mypic1", "height": 10, "width": 20},
                {"name": "mypic2", "height": 30, "width": 40},
            ],
            "tables": [
                {
                    "name": "Table1",
                    "range_address": "A10:B11",
                    "header_row_range_address": "A10:B10",
                    "data_body_range_address": "A11:B11",
                    "total_row_range_address": None,
                    "show_headers": True,
                    "show_totals": False,
                    "table_style": "TableStyleMedium2",
                    "show_autofilter": True,
                },
                {
                    "name": "Table2",
                    "range_address": "A15:C17",
                    "header_row_range_address": None,
                    "data_body_range_address": "A15:C16",
                    "total_row_range_address": "A17:C17",
                    "show_headers": False,
                    "show_totals": True,
                    "table_style": "TableStyleLight1",
                    "show_autofilter": False,
                },
            ],
        },
        {
            "name": "Sheet2",
            "values": [["aa", "bb"], [11.1, 22.2]],
            "pictures": [],
            "tables": [],
        },
        {
            "name": "Sheet3",
            "values": [
                ["", "string"],
                [-1.1, 1.1],
                [True, False],
                ["2021-10-01T00:00:00.000Z", 44561.9826388889],
            ],
            "pictures": [],
            "tables": [],
        },
    ],
}

expected_body["VBA"] = {
    "client": "VBA",
    "version": "dev",
    "book": {"name": "engines.xlsm", "active_sheet_index": 0, "selection": "A1"},
    "names": [
        {
            "name": "one",
            "sheet_index": 0,
            "address": "A1",
            "book_scope": True,
            "scope_sheet_name": None,
            "scope_sheet_index": None,
        },
        {
            "name": "'Sheet 1'!two",
            "sheet_index": 0,
            "address": "C7:D8",
            "book_scope": False,
            "scope_sheet_name": "Sheet 1",
            "scope_sheet_index": 0,
        },
        {
            "name": "Sheet2!two",
            "sheet_index": 2,
            "address": "B3",
            "book_scope": False,
            "scope_sheet_name": "Sheet2",
            "scope_sheet_index": 1,
        },
        {
            "name": "two",
            "sheet_index": 1,
            "address": "A1:A2",
            "book_scope": True,
            "scope_sheet_name": None,
            "scope_sheet_index": None,
        },
    ],
    "sheets": [
        {
            "name": "Sheet 1",
            "pictures": [
                {"name": "mypic1", "height": 10, "width": 20},
                {"name": "mypic2", "height": 30, "width": 40},
            ],
            "tables": [
                {
                    "name": "Table1",
                    "range_address": "$A$10:$B$11",
                    "header_row_range_address": "$A$10:$B$10",
                    "data_body_range_address": "$A$11:$B$11",
                    "total_row_range_address": None,
                    "show_headers": True,
                    "show_totals": False,
                    "table_style": "TableStyleMedium2",
                    "show_autofilter": True,
                },
                {
                    "name": "Table2",
                    "range_address": "$A$15:$C$17",
                    "header_row_range_address": None,
                    "data_body_range_address": "$A$15:$C$16",
                    "total_row_range_address": "$A$17:$C$17",
                    "show_headers": False,
                    "show_totals": True,
                    "table_style": "TableStyleLight1",
                    "show_autofilter": False,
                },
            ],
            "values": [
                ["a", "b", "c", None],
                [1.1, 2.2, 3.3, "2021-01-01T00:00:00.000Z"],
                [4.4, 5.5, 6.6, None],
                [None, None, None, None],
                [None, None, None, None],
                [None, None, None, None],
                [None, None, None, None],
                [None, None, None, None],
                [None, None, None, None],
                ["Column1", "Column2", None, None],
                [1.1, 2.2, None, None],
                [None, None, None, None],
                [None, None, None, None],
                [None, None, None, None],
                [1.1, 2.2, 3.3, None],
                [4.4, 5.5, 6.6, None],
                ["Total", None, 9.9, None],
            ],
        },
        {
            "name": "Sheet2",
            "pictures": [],
            "tables": [],
            "values": [["aa", "bb"], [11.1, 22.2]],
        },
        {
            "name": "Sheet3",
            "pictures": [],
            "tables": [],
            "values": [
                [None, "string"],
                [-1.1, 1.1],
                [True, False],
                ["2021-10-01T00:00:00.000Z", "2021-12-31T23:35:00.000Z"],
            ],
        },
    ],
}
expected_body["Office Scripts"] = {
    "client": "Microsoft Office Scripts",
    "version": "dev",
    "book": {"name": "engines.xlsm", "active_sheet_index": 0, "selection": "A1"},
    "names": [
        {
            "name": "one",
            "sheet_index": 0,
            "address": "A1",
            "scope_sheet_name": None,
            "scope_sheet_index": None,
            "book_scope": True,
        },
        {
            "name": "two",
            "sheet_index": 1,
            "address": "A1:A2",
            "scope_sheet_name": None,
            "scope_sheet_index": None,
            "book_scope": True,
        },
        {
            "name": "two",
            "sheet_index": 0,
            "address": "C7:D8",
            "scope_sheet_name": "Sheet 1",
            "scope_sheet_index": 0,
            "book_scope": False,
        },
        {
            "name": "two",
            "sheet_index": 2,
            "address": "B3",
            "scope_sheet_name": "Sheet2",
            "scope_sheet_index": 1,
            "book_scope": False,
        },
    ],
    "sheets": [
        {
            "name": "Sheet 1",
            "values": [
                ["a", "b", "c", ""],
                [1.1, 2.2, 3.3, "2021-01-01T00:00:00.000Z"],
                [4.4, 5.5, 6.6, ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["Column1", "Column2", "", ""],
                [1.1, 2.2, "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                [1.1, 2.2, 3.3, ""],
                [4.4, 5.5, 6.6, ""],
                ["Total", "", 9.9, ""],
            ],
            # Width/height differ for Desktop and Web
            "pictures": [
                {"name": "mypic1", "width": 20, "height": 10},
                {"name": "mypic2", "width": 40, "height": 30},
            ],
            "tables": [
                {
                    "name": "Table1",
                    "range_address": "A10:B11",
                    "header_row_range_address": "A10:B10",
                    "data_body_range_address": "A11:B11",
                    "total_row_range_address": None,
                    "show_headers": True,
                    "show_totals": False,
                    "table_style": "TableStyleMedium2",
                    "show_autofilter": True,
                },
                {
                    "name": "Table2",
                    "range_address": "A15:C17",
                    "header_row_range_address": None,
                    "data_body_range_address": "A15:C16",
                    "total_row_range_address": "A17:C17",
                    "show_headers": False,
                    "show_totals": True,
                    "table_style": "TableStyleLight1",
                    "show_autofilter": False,
                },
            ],
        },
        {
            "name": "Sheet2",
            "values": [["aa", "bb"], [11.1, 22.2]],
            "pictures": [],
            "tables": [],
        },
        {
            "name": "Sheet3",
            "values": [
                ["", "string"],
                [-1.1, 1.1],
                [True, False],
                ["2021-10-01T00:00:00.000Z", 44561.9826388889],
            ],
            "pictures": [],
            "tables": [],
        },
    ],
}
expected_body["Google Apps Script"] = {
    "client": "Google Apps Script",
    "version": "dev",
    "book": {"name": "engines.xlsm", "active_sheet_index": 0, "selection": "A1"},
    "names": [
        {
            "name": "one",
            "sheet_index": 0,
            "address": "A1",
            "scope_sheet_name": None,
            "scope_sheet_index": None,
            "book_scope": True,
        },
        {
            "name": "'Sheet 1'!two",
            "sheet_index": 0,
            "address": "C7:D8",
            "scope_sheet_name": "Sheet 1",
            "scope_sheet_index": 0,
            "book_scope": False,
        },
        {
            "name": "Sheet2!two",
            "sheet_index": 2,
            "address": "B3",
            "scope_sheet_name": "Sheet3",
            "scope_sheet_index": 2,
            "book_scope": False,
        },
        {
            "name": "two",
            "sheet_index": 1,
            "address": "A1:A2",
            "scope_sheet_name": None,
            "scope_sheet_index": None,
            "book_scope": True,
        },
    ],
    "sheets": [
        {
            "name": "Sheet 1",
            "values": [
                ["a", "b", "c", ""],
                [1.1, 2.2, 3.3, "2021-01-01T00:00:00.000Z"],
                [4.4, 5.5, 6.6, ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["Column1", "Column2", "", ""],
                [1.1, 2.2, "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                ["", "", "", ""],
                [1.1, 2.2, 3.3, ""],
                [4.4, 5.5, 6.6, ""],
                ["Total", "", 9.899999999999999, ""],
            ],
            "pictures": [
                {"name": "", "height": 13, "width": 35},
                {"name": "", "height": 40, "width": 62},
            ],
            "tables": [],
        },
        {
            "name": "Sheet2",
            "values": [["aa", "bb"], [11.1, 22.2]],
            "pictures": [],
            "tables": [],
        },
        {
            "name": "Sheet3",
            "values": [
                ["", "string"],
                [-1.1, 1.1],
                [True, False],
                ["2021-10-01T00:00:00.000Z", "2021-12-31T23:35:00.000Z"],
            ],
            "pictures": [],
            "tables": [],
        },
    ],
}
if __name__ == "__main__":
    import uvicorn

    uvicorn.run(
        "devserver:app",
        host="127.0.0.1",
        port=8000,
        reload=True,
        reload_dirs=[this_dir, this_dir.parent / "xlwings"],
        ssl_keyfile=this_dir / "localhost+2-key.pem",
        ssl_certfile=this_dir / "localhost+2.pem",
    )



from pathlib import Path

import pytest

import xlwings as xw

this_dir = Path(__file__).resolve().parent


@pytest.fixture(scope="module")
def app():
    with xw.App(visible=False) as app:
        yield app

    for f in Path(".").glob("tempfile*"):
        f.unlink()
    for f in Path("temp").glob("tempfile*"):
        f.unlink()


def test_save_new_book_defaults(app):
    book = app.books.add()
    if Path(book.name + ".xlsx").is_file():
        Path(book.name + ".xlsx").unlink()
    book.save()

    assert Path(book.name).is_file()


# TODO: xlam and xltx fail


@pytest.mark.parametrize(
    "name",
    [
        "tempfile.xlsx",
        "tempfile.xlsm",
        "tempfile.xlsb",
        "tempfile.xltm",
        "tempfile.xls",
        "tempfile.xlt",
        "tempfile.xla",
    ],
)
def test_save_new_book_no_path(app, name):
    book = app.books.add()
    book.save(name)
    assert book.name == name
    assert Path(name).is_file()


@pytest.mark.parametrize(
    "name",
    [
        "tempfile2.xlsx",
        "tempfile2.xlsm",
        "tempfile2.xlsb",
        "tempfile2.xltm",
        "tempfile2.xls",
        "tempfile2.xlt",
        "tempfile2.xla",
    ],
)
def test_save_new_book_with_path(app, name):
    Path("temp").mkdir(exist_ok=True)
    book = app.books.add()
    fullname = Path(".").resolve() / "temp" / name
    book.save(fullname)
    assert book.fullname == str(fullname)
    assert Path(fullname).is_file()


@pytest.mark.parametrize(
    "name",
    [
        "tempfile3.xlsx",
        "tempfile3.xlsm",
        "tempfile3.xlsb",
        "tempfile3.xltm",
        "tempfile3.xls",
        "tempfile3.xlt",
        "tempfile3.xla",
    ],
)
def test_save_existing_book_no_path(app, name):
    book = app.books.open(this_dir / "test book.xlsx")
    book.save(name)
    book.save()
    assert book.name == name
    assert Path(name).is_file()


@pytest.mark.parametrize(
    "name",
    [
        "tempfile4.xlsx",
        "tempfile4.xlsm",
        "tempfile4.xlsb",
        "tempfile4.xltm",
        "tempfile4.xls",
        "tempfile4.xlt",
        "tempfile4.xla",
    ],
)
def test_save_existing_book_with_path(app, name):
    Path("temp").mkdir(exist_ok=True)
    book = app.books.open(this_dir / "test book.xlsx")
    fullname = Path(".").resolve() / "temp" / name
    book.save(fullname)
    book.save()
    assert book.fullname == str(fullname)
    assert Path(fullname).is_file()
